#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import logging
import numpy as np
import scipy.optimize as sciopt
import itertools

from TransportMaps.Misc import generate_total_order_midxs, mpi_map, mpi_alloc_dmem, \
    cached, counted, deprecate
from TransportMaps.Functionals.ParametricFunctionApproximationBase import *

__all__ = ['LinearSpanApproximation']

nax = np.newaxis

class LinearSpanApproximation(TensorizedFunctionApproximation):
    r""" Parametric function :math:`f_{\bf a} = \sum_{{\bf i} \in \mathcal{I}} {\bf a}_{\bf i} \Phi_{\bf i}`

    Args:
      basis_list (list): list of :math:`d`
        :class:`OrthogonalBasis<SpectralToolbox.OrthogonalBasis>`
      spantype (str): Span type. 'total' total order, 'full' full order,
        'midx' multi-indeces specified
      order_list (:class:`list<list>` of :class:`int<int>`): list of 
        orders :math:`\{N_i\}_{i=0}^d`
      multi_idxs (list): list of tuples containing the active multi-indices
      full_basis_list (list): full list of :class:`Basis<SpectralToolbox.Basis>`.
        ``basis_list`` is a subset of ``full_basis_list``. This may be used to
        automatically increase the input dimension of the approximation.
    """

    def __init__(self, basis_list, spantype=None,
                 order_list=None, multi_idxs=None,
                 full_basis_list=None):
        if spantype in ['total','full'] and order_list is not None:
            self.max_order_list = order_list
            self.multi_idxs = self.generate_multi_idxs(spantype)
        elif spantype == 'midx' and multi_idxs is not None:
            self.multi_idxs = multi_idxs
            self.max_order_list = list( np.max(np.asarray(self.multi_idxs), axis=0) )
        else:
            raise ValueError("""
                             Parameters mismatch: (spantype==total|full & order_list != None) |
                             (spantype==midx & multi_idxs != None)
                             """)
        super(LinearSpanApproximation,self).__init__(basis_list, full_basis_list)

    def generate_multi_idxs(self, spantype):
        r""" Generate the list of multi-indices
        """
        if spantype == 'full':
            return list(itertools.product(*[range(o+1) for o in self.max_order_list]))
        elif spantype == 'total':
            midxs = generate_total_order_midxs(self.max_order_list)
            return midxs
        raise NotImplementedError("Not implemented for the selected spantype (%s)" % spantype)

    def init_coeffs(self):
        r""" Initialize the coefficients :math:`{\bf a}`
        """
        self._coeffs = np.zeros(self.n_coeffs)

    @property
    def n_coeffs(self):
        r""" Get the number :math:`N` of coefficients :math:`{\bf a}`

        Returns:
          (:class:`int<int>`) -- number of coefficients
        """
        return len(self.multi_idxs)
        
    @property
    def coeffs(self):
        r""" Get the coefficients :math:`{\bf a}`

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`N`]) -- coefficients
        """
        return self._coeffs

    @coeffs.setter
    def coeffs(self, coeffs):
        r""" Set the coefficients :math:`{\bf a}`.

        Args:
          coeffs (:class:`ndarray<numpy.ndarray>` [:math:`N`]): coefficients
        """
        if len(coeffs) != self.n_coeffs:
            raise ValueError("The number of input coefficients does not agree " +
                             "with the number of expected coefficients.")
        self._coeffs = coeffs

    @deprecate("LinearSpanApproximation.get_multi_idxs()",
               "2.0",
               "Use property LinearSpanApproximation.multi_idxs")
    def get_multi_idxs(self):
        r""" Get the list of multi indices

        Return:
          (:class:`list` of :class:`tuple`) -- multi indices
        """
        return self.multi_idxs

    @deprecate("LinearSpanApproximation.set_multi_idxs()",
               "2.0",
               "Use property LinearSpanApproximation.multi_idxs")
    def set_multi_idxs(self, multi_idxs):
        r""" Set the list of multi indices

        Args:
          multi_idxs (:class:`list` of :class:`tuple`): multi indices
        """
        self.multi_idxs = multi_idxs

    @property
    def multi_idxs(self):
        return self._multi_idxs[:]

    @multi_idxs.setter
    def multi_idxs(self, midxs):
        self._multi_idxs = midxs
        self.max_order_list = list( np.max(np.asarray(midxs), axis=0) )

    def get_directional_orders(self):
        r""" Get the maximum orders of the univariate part of the representation.

        Returns:
          (:class:`list<list>` [d] :class:`int<int>`) -- orders
        """
        if not hasattr(self, "max_order_list"): # Backcompatibility
            self.max_order_list = list( np.max(np.asarray(self.multi_idxs), axis=0) )
        return self.max_order_list[:]

    def precomp_Vandermonde_evaluate(self, x, precomp=None):
        r""" Precompute the multi-variate Vandermonde matrices for the evaluation of :math:`f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary containing the Vandermonde matrix
        """
        if precomp is None: precomp = {}
        try: V = precomp['V']
        except KeyError as e:
            try:
                V_list = precomp['V_list']
            except (TypeError, KeyError) as e:
                self.precomp_evaluate(x, precomp)
                V_list = precomp['V_list']
            precomp['V'] = np.ones((V_list[0].shape[0], self.n_coeffs))
            for i,midx in enumerate(self.multi_idxs):
                for idx,V1d in zip(midx,V_list):
                    precomp['V'][:,i] *= V1d[:,idx]
        return precomp
        
    @cached()
    @counted
    def evaluate(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- function evaluations
        """
        try:
            V = precomp['V']
        except (TypeError, KeyError) as e:
            try:
                V_list = precomp['V_list']
            except (TypeError, KeyError) as e:
                precomp = self.precomp_evaluate(x, precomp)
                idxs_slice = slice(None)
                V_list = precomp['V_list']
            tot_size = V_list[0].shape[0]
            out = np.zeros(x.shape[0])
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate this loop?
            for c, midx in zip(self._coeffs, self.multi_idxs):
                tmp[:] = 1.
                for idx,V in zip(midx,V_list):
                    tmp *= V[idxs_slice,idx]
                out += c * tmp
        else:
            tot_size = V.shape[0]
            out = np.dot(V[idxs_slice,:], self._coeffs)
        return out

    def precomp_Vandermonde_grad_x(self, x, precomp=None):
        r""" Precompute the multi-variate Vandermonde matrices for the evaluation of :math:`\nabla_{\bf x} f_{\bf a}` at ``x``

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Return:
          (:class:`dict<dict>` with :class:`list<list>`
            [:math:`d`] of :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary containing the list of multi-variate Vandermonde matrices.
        """
        if precomp is None: precomp = {}
        try: grad_x_V_list = precomp['grad_x_V_list']
        except KeyError as e:
            try: V_list = precomp['V_list']
            except KeyError as e:
                self.precomp_evaluate(x, precomp)
                V_list = precomp['V_list']
            try: partial_x_V_list = precomp['partial_x_V_list']
            except KeyError as e:
                self.precomp_grad_x(x, precomp)
                partial_x_V_list = precomp['partial_x_V_list']
            grad_x_V_list = []
            # TODO: Accelerate this loops?
            for d in range(self.dim):
                grad_x_V = np.ones((x.shape[0], self.n_coeffs))
                for i,midx in enumerate(self.multi_idxs):
                    for j, (idx, V1d, pxV1d) in enumerate(zip(midx, V_list, partial_x_V_list)):
                        if j != d:
                            grad_x_V[:,i] *= V1d[:,idx]
                        else:
                            grad_x_V[:,i] *= pxV1d[:,idx]
                grad_x_V_list.append( grad_x_V )
            precomp['grad_x_V_list'] = grad_x_V_list
        return precomp

    @cached()
    @counted
    def grad_x(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla_{\bf x} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x} f_{\bf a}({\bf x})`
        """
        try:
            grad_x_V_list = precomp['grad_x_V_list']
        except (TypeError, KeyError) as e:
            try:
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
            except (TypeError,KeyError) as e:
                # Clean pre-existing
                precomp.pop('V_list', None)
                precomp.pop('partial_x_V_list', None)
                # Ignoring slice
                idxs_slice = slice(None)
                # Compute
                precomp = self.precomp_evaluate(x, precomp)
                precomp = self.precomp_grad_x(x, precomp)
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
            out = np.zeros( x.shape )
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate this loops?
            for d in range(self.dim):
                for c, midx in zip(self._coeffs, self.multi_idxs):
                    tmp[:] = 1.
                    for j, (idx, V1d, pxV1d) in enumerate(zip(midx, V_list,
                                                              partial_x_V_list)):
                        if j != d:
                            tmp *= V1d[idxs_slice,idx]
                        else:
                            tmp *= pxV1d[idxs_slice,idx]
                    out[:,d] += c * tmp
        else:
            out = np.zeros( x.shape )
            for i, grad_x_V in enumerate(grad_x_V_list):
                out[idxs_slice,i] = np.dot( grad_x_V[idxs_slice,:], self._coeffs )
        return out

    @cached()
    @counted
    def grad_a_grad_x(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla_{\bf a} \nabla_{\bf x} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N,d`]) --
            :math:`\nabla_{\bf a} \nabla_{\bf x} f_{\bf a}({\bf x})`
        """
        try:
            grad_x_V_list = precomp['grad_x_V_list']
        except (TypeError, KeyError) as e:
            try: V_list = precomp['V_list']
            except KeyError as e:
                precomp = self.precomp_evaluate(x)
            except TypeError as e:
                x = x[idxs_slice,:]
                precomp = self.precomp_evaluate(x)
                idxs_slice = slice(None)
            finally:
                V_list = precomp['V_list']
            try: partial_x_V_list = precomp['partial_x_V_list']
            except KeyError as e:
                self.precomp_grad_x(x, precomp)
                partial_x_V_list = precomp['partial_x_V_list']
            out = np.zeros( (x.shape[0], self.n_coeffs, x.shape[1]) )
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate this loops?
            for d in range(self.dim):
                cidx = 0
                for c, midx in zip(self._coeffs, self.multi_idxs):
                    tmp[:] = 1.
                    for j, (idx, V1d, pxV1d) in enumerate(zip(midx, V_list,
                                                              partial_x_V_list)):
                        if j != d:
                            tmp *= V1d[idxs_slice,idx]
                        else:
                            tmp *= pxV1d[idxs_slice,idx]
                    out[:,cidx,d] += tmp#*(c != 0)
                    cidx = cidx + 1
        else:
            out = np.zeros( (x.shape[0], self.n_coeffs, x.shape[1]) )
            # coeff_ind = indicator function applied to coefficients in map component
            #coeff_ind = [int(x != 0) for x in self._coeffs]
            for i, grad_x_V in enumerate(grad_x_V_list):
                out[idxs_slice,:,i] = grad_x_V[idxs_slice,:]#np.dot(grad_x_V[idxs_slice,:], coeff_ind)
        return out

    def precomp_Vandermonde_hess_x(self, x, precomp=None):
        r""" Precompute the multi-variate Vandermonde matrices for the evaluation of :math:`\nabla^2_{\bf x} f_{\bf a}` at ``x``

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Return:
          (:class:`dict<dict>` with :class:`ndarray<numpy.ndarray>`
            [:math:`d,d`] of :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary containing the matrix of multi-variate Vandermonde matrices
        """
        if precomp is None: precomp = {}
        try: hess_x_V_mat = precomp['hess_x_V_mat']
        except KeyError as e:
            try: V_list = precomp['V_list']
            except KeyError as e:
                self.precomp_evaluate(x, precomp)
                V_list = precomp['V_list']
            try: partial_x_V_list = precomp['partial_x_V_list']
            except KeyError as e:
                self.precomp_grad_x(x, precomp)
                partial_x_V_list = precomp['partial_x_V_list']
            try: partial2_x_V_list = precomp['partial2_x_V_list']
            except KeyError as e:
                self.precomp_hess_x(x, precomp)
                partial2_x_V_list = precomp['partial2_x_V_list']
            hess_x_V_mat = np.empty((self.dim,self.dim), dtype=object)
            # TODO: Accelerate these loops?
            for d1 in range(self.dim):
                for d2 in range(self.dim):
                    hess_x_V = np.ones((x.shape[0],self.n_coeffs))
                    for i, midx in enumerate(self.multi_idxs):
                        for j, (idx, V1d, pxV1d, p2xV1d) in enumerate(zip(
                                midx, V_list, partial_x_V_list, partial2_x_V_list)):
                            if d1 == d2 and j == d1:
                                hess_x_V[:,i] *= p2xV1d[:,idx]
                            elif j == d1 or j == d2:
                                hess_x_V[:,i] *= pxV1d[:,idx]
                            else:
                                hess_x_V[:,i] *= V1d[:,idx]
                    hess_x_V_mat[d1,d2] = hess_x_V
            precomp['hess_x_V_mat'] = hess_x_V_mat
        return precomp

    @cached()
    @counted
    def hess_x(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla^2_{\bf x} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) --
            :math:`\nabla^2_{\bf x} f_{\bf a}({\bf x})`
        """
        try:
            hess_x_V_mat = precomp['hess_x_V_mat']
        except (TypeError, KeyError) as e:
            try:
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
                partial2_x_V_list = precomp['partial2_x_V_list']
            except (TypeError,KeyError) as e:
                # Clean pre-existing
                precomp.pop('V_list', None)
                precomp.pop('partial_x_V_list', None)
                precomp.pop('partial2_x_V_list', None)
                # Ignoring slice
                idxs_slice = slice(None)
                # Compute
                precomp = self.precomp_evaluate(x, precomp)
                precomp = self.precomp_grad_x(x, precomp)
                precomp = self.precomp_hess_x(x, precomp)
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
                partial2_x_V_list = precomp['partial2_x_V_list']
            out = np.zeros((x.shape[0],self.dim,self.dim))
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate these loops?
            for d1 in range(self.dim):
                for d2 in range(self.dim):
                    for i, (c, midx) in enumerate(zip(self._coeffs,self.multi_idxs)):
                        tmp[:] = 1.
                        for j, (idx, V1d,
                                pxV1d, p2xV1d) in enumerate(zip(midx, V_list,
                                                                partial_x_V_list,
                                                                partial2_x_V_list)):
                            if d1 == d2 and j == d1:
                                tmp *= p2xV1d[idxs_slice,idx]
                            elif j == d1 or j == d2:
                                tmp *= pxV1d[idxs_slice,idx]
                            else:
                                tmp *= V1d[idxs_slice,idx]
                        out[:,d1,d2] += c * tmp
        else:
            out = np.zeros((x.shape[0],self.dim,self.dim))
            for i in range(self.dim):
                for j in range(self.dim):
                    out[idxs_slice,i,j] = np.dot( hess_x_V_mat[i,j][idxs_slice,:],
                                                  self._coeffs )
        return out

    @cached()
    @counted
    def grad_a_hess_x(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla_{\bf a} \nabla^2_{\bf x} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N,d,d`]) --
            :math:`\nabla_{\bf a} \nabla^2_{\bf x} f_{\bf a}({\bf x})`
        """
        try:
            hess_x_V_mat = precomp['hess_x_V_mat']
        except (TypeError, KeyError) as e:
            try: V_list = precomp['V_list']
            except KeyError as e:
                precomp = self.precomp_evaluate(x)
            except TypeError as e:
                x = x[idxs_slice,:]
                precomp = self.precomp_evaluate(x)
                idxs_slice = slice(None)
            finally:
                V_list = precomp['V_list']
            try: partial_x_V_list = precomp['partial_x_V_list']
            except KeyError as e:
                self.precomp_grad_x(x, precomp)
                partial_x_V_list = precomp['partial_x_V_list']
            try: partial2_x_V_list = precomp['partial2_x_V_list']
            except KeyError as e:
                self.precomp_hess_x(x, precomp)
                partial2_x_V_list = precomp['partial2_x_V_list']
            out = np.zeros((x.shape[0], self.n_coeffs, self.dim,self.dim))
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate these loops?
            for d1 in range(self.dim):
                for d2 in range(self.dim):
                    cidx = 0
                    for i, (c, midx) in enumerate(zip(self._coeffs,self.multi_idxs)):
                        tmp[:] = 1.
                        for j, (idx, V1d,
                                pxV1d, p2xV1d) in enumerate(zip(midx, V_list,
                                                                partial_x_V_list,
                                                                partial2_x_V_list)):
                            if d1 == d2 and j == d1:
                                tmp *= p2xV1d[idxs_slice,idx]
                            elif j == d1 or j == d2:
                                tmp *= pxV1d[idxs_slice,idx]
                            else:
                                tmp *= V1d[idxs_slice,idx]
                        out[:,cidx,d1,d2] += tmp#*(c != 0)
                        cidx = cidx + 1
        else:
            out = np.zeros((x.shape[0], self.n_coeffs, self.dim,self.dim))
            #coeff_ind = [int(x != 0) for x in self._coeffs]
            for i in range(self.dim):
                for j in range(self.dim):
                    out[idxs_slice,:,i,j] = hess_x_V_mat[i,j][idxs_slice,:]#np.dot(hess_x_V_mat[i,j][idxs_slice,:], coeff_ind)
        return out

    @cached()
    @counted
    def grad_a(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla_{\bf a} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            :math:`\nabla_{\bf a} f_{\bf a}({\bf x})`
        """
        try:
            V = precomp['V']
        except KeyError as e:
            if 'V_list' not in precomp:
                idxs_slice = slice(None)
            precomp = self.precomp_Vandermonde_evaluate(x, precomp)
        except TypeError as e:
            idxs_slice = slice(None)
            precomp = self.precomp_Vandermonde_evaluate(x, precomp)
        finally:
            V = precomp['V']
        return V[idxs_slice,:]

    @cached()
    @counted
    def grad_a_t(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\left(\nabla_{\bf a} f_{\bf a}\right)^T` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            :math:`\left(\nabla_{\bf a} f_{\bf a}({\bf x}\right)^T)`
        """
        try:
            VT = precomp['VT']
        except (TypeError, KeyError) as e:
            try:
                V = precomp['V']
            except KeyError:
                if 'V_list' not in precomp:
                    idxs_slice = slice(None)
                precomp = self.precomp_Vandermonde_evaluate(x, precomp)
            except TypeError:
                idxs_slice = slice(None)
                precomp = self.precomp_Vandermonde_evaluate(x, precomp)
            finally:
                V = precomp['V']
            VT = np.transpose(V).copy()
            precomp['VT'] = VT
        return VT[:,idxs_slice]

    def precomp_VVT_evaluate(self, x, precomp=None):
        r""" Precompute the product :math:`VV^T` of the multi-variate Vandermonde matrices for the evaluation of :math:`f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary containing the desired product
        """
        try:
            V = precomp['V']
        except (TypeError, KeyError) as e:
            precomp = self.precomp_Vandermonde_evaluate(x, precomp)
            V = precomp['V']
        precomp['VVT'] = V[:,:,nax] * V[:,nax,:]
        return precomp

    @cached()
    @counted
    def grad_a_squared(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\left(\nabla_{\bf a} f_{\bf a}\right)\left(\nabla_{\bf a} f_{\bf a}\right)^T` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N,N`]) --
            :math:`\left(\nabla_{\bf a} f_{\bf a}\right)\left(\nabla_{\bf a} f_{\bf a}\right)^T`
        """
        try:
            VVT = precomp['VVT']
        except KeyError:
            if 'V_list' not in precomp:
                idxs_slice = slice(None)
            precomp = self.precomp_VVT_evaluate(x, precomp)
        except TypeError:
            idxs_slice = slice(None)
            precomp = self.precomp_VVT_evaluate(x, precomp)
        finally:
            VVT = precomp['VVT']
        return VVT[idxs_slice,:,:]

    @counted
    def hess_a(self, x, precomp=None, idxs_slice=slice(None), *arg, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf a} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N,N`]) --
            :math:`\nabla^2_{\bf a} f_{\bf a}({\bf x})`
        """
        return np.zeros((1,self.n_coeffs,self.n_coeffs))

    def precomp_Vandermonde_partial_xd(self, x, precomp=None):
        r""" Precompute multi-variate Vandermonde matrix for the evaluation of :math:`\partial_{x_d} f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary with Vandermonde matrix
        """
        if precomp is None: precomp = {}
        try: partial_xd_V = precomp['partial_xd_V']
        except KeyError as e:
            try: V_list = precomp['V_list']
            except KeyError as e:
                self.precomp_evaluate(x, precomp)
                V_list = precomp['V_list']
            try: partial_xd_V_last = precomp['partial_xd_V_last']
            except KeyError as e:
                self.precomp_partial_xd(x, precomp)
                partial_xd_V_last = precomp['partial_xd_V_last']
            partial_xd_V = np.ones((V_list[0].shape[0], self.n_coeffs))
            for i, midx in enumerate(self.multi_idxs):
                for idx, V1d in zip(midx[:-1], V_list[:-1]):
                    partial_xd_V[:,i] *= V1d[:,idx]
                partial_xd_V[:,i] *= partial_xd_V_last[:,midx[-1]]
            precomp['partial_xd_V'] = partial_xd_V
        return precomp

    @cached()
    @counted
    def partial_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial_{x_d} f_{\bf a}({\bf x})`
        """
        try:
            partial_xd_V = precomp['partial_xd_V']
        except (TypeError, KeyError) as e:
            try:
                V_list = precomp['V_list']
                partial_xd_V_last = precomp['partial_xd_V_last']
            except (TypeError, KeyError) as e:
                # Clean pre-existing
                precomp.pop('V_list', None)
                precomp.pop('partial_x_V_last', None)
                # Ignoring slice
                idxs_slice = slice(None)
                # Compute
                precomp = self.precomp_evaluate(x, precomp)
                precomp = self.precomp_partial_xd(x, precomp)
                V_list = precomp['V_list']
                partial_xd_V_last = precomp['partial_xd_V_last']
            tot_size = V_list[0].shape[0]
            out = np.zeros(x.shape[0])
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate these loops?
            for i, (c, midx) in enumerate(zip(self._coeffs,self.multi_idxs)):
                tmp[:] = 1.
                for idx, V1d in zip(midx[:-1], V_list[:-1]):
                    tmp *= V1d[idxs_slice,idx]
                tmp *= partial_xd_V_last[idxs_slice,midx[-1]]
                out += c * tmp
        else:
            tot_size = partial_xd_V.shape[0]
            out = np.dot(partial_xd_V[idxs_slice,:], self._coeffs)
        return out
    
    def precomp_Vandermonde_grad_x_partial_xd(self, x, precomp=None):
        r""" Precompute multi-variate Vandermonde matrices for the evaluation of :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`list<list>` [d]
            :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary containing the list of multi-variate Vandermonde matrices.
        """
        if precomp is None: precomp = {}
        try: grad_x_partial_xd_V_list = precomp['grad_x_partial_xd_V_list']
        except KeyError as e:
            try: V_list = precomp['V_list']
            except KeyError as e:
                self.precomp_evaluate(x, precomp)
                V_list = precomp['V_list']
            try: partial_x_V_list = precomp['partial_x_V_list']
            except KeyError as e:
                self.precomp_grad_x(x, precomp)
                partial_x_V_list = precomp['partial_x_V_list']
            try: partial2_xd_V_last = precomp['partial2_xd_V_last']
            except KeyError as e:
                self.precomp_partial2_xd(x, precomp)
                partial2_xd_V_last = precomp['partial2_xd_V_last']
            grad_x_partial_xd_V_list = []
            # TODO: Accelerate these loops?
            for d in range(self.dim):
                grad_x_partial_xd_V = np.ones((x.shape[0], self.n_coeffs))
                for i, midx in enumerate(self.multi_idxs):
                    for j, (idx, V1d, pxV1d) in enumerate(zip(midx, V_list,
                                                              partial_x_V_list)):
                        if j == d and d == (self.dim-1):
                            grad_x_partial_xd_V[:,i] *= partial2_xd_V_last[:,idx]
                        elif j == d or j == (self.dim-1):
                            grad_x_partial_xd_V[:,i] *= pxV1d[:,idx]
                        else:
                            grad_x_partial_xd_V[:,i] *= V1d[:,idx]
                grad_x_partial_xd_V_list.append( grad_x_partial_xd_V )
            precomp['grad_x_partial_xd_V_list'] = grad_x_partial_xd_V_list
        return precomp

    @cached()
    @counted
    def grad_x_partial_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        try:
            grad_x_partial_xd_V_list = precomp['grad_x_partial_xd_V_list']
        except (TypeError, KeyError) as e:
            try:
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
                partial2_xd_V_last = precomp['partial2_xd_V_last']
            except (TypeError, KeyError) as e:
                # Clean pre-existing
                precomp.pop('V_list', None)
                precomp.pop('partial_x_V_list', None)
                precomp.pop('partial2_xd_V_last', None)
                # Ignoring slice
                idxs_slice = slice(None)
                # Compute
                precomp = self.precomp_evaluate(x, precomp)
                precomp = self.precomp_grad_x(x, precomp)
                precomp = self.precomp_partial2_xd(x, precomp)
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
                partial2_xd_V_last = precomp['partial2_xd_V_last']
            out = np.zeros( x.shape )
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate these loops?
            for d in range(self.dim):
                for i, (c, midx) in enumerate(zip(self._coeffs,self.multi_idxs)):
                    tmp[:] = 1.
                    for j, (idx, V1d, pxV1d) in enumerate(zip(midx, V_list,
                                                              partial_x_V_list)):
                        if j == d and d == (self.dim-1):
                            tmp *= partial2_xd_V_last[:,idx]
                        elif j == d or j == (self.dim-1):
                            tmp *= pxV1d[:,idx]
                        else:
                            tmp *= V1d[:,idx]
                    out[:,d] += c * tmp
        else:
            out = np.zeros( x.shape )
            for i in range( self.dim ):
                out[:,i] = np.dot(
                    grad_x_partial_xd_V_list[i][idxs_slice,:], self._coeffs )
        return out

    @cached()
    @counted
    def grad_a_grad_x_partial_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla_{\bf a}\nabla_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N,d`]) --
            :math:`\nabla_{\bf a}\nabla_{\bf x}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        try:
            grad_x_partial_xd_V_list = precomp['grad_x_partial_xd_V_list']
        except (TypeError, KeyError) as e:
            try: V_list = precomp['V_list']
            except (TypeError, KeyError) as e:
                precomp = self.precomp_evaluate(x, precomp)
                V_list = precomp['V_list']
            try: partial_x_V_list = precomp['partial_x_V_list']
            except (TypeError, KeyError) as e:
                precomp = self.precomp_grad_x(x, precomp)
                partial_x_V_list = precomp['partial_x_V_list']
            try: partial2_xd_V_last = precomp['partial2_xd_V_last']
            except (TypeError, KeyError) as e:
                precomp = self.precomp_partial2_xd(x, precomp)
                partial2_xd_V_last = precomp['partial2_xd_V_last']
            out = np.zeros((x.shape[0], self.n_coeffs, self.dim))
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate these loops?
            for d in range(self.dim):
                cidx = 0
                for i, (c, midx) in enumerate(zip(self._coeffs,self.multi_idxs)):
                    tmp[:] = 1.
                    for j, (idx, V1d, pxV1d) in enumerate(zip(midx, V_list,
                                                              partial_x_V_list)):
                        if j == d and d == (self.dim-1):
                            tmp *= partial2_xd_V_last[:,idx]
                        elif j == d or j == (self.dim-1):
                            tmp *= pxV1d[:,idx]
                        else:
                            tmp *= V1d[:,idx]
                    out[:,cidx,d] += tmp#*(c != 0)
                    cidx = cidx + 1
        else:
            out = np.zeros((x.shape[0], self.n_coeffs, self.dim))
            #coeff_ind = [int(x != 0) for x in self._coeffs]
            for i in range( self.dim ):
                out[:,:,i] = grad_x_partial_xd_V_list[i]#np.dot(grad_x_partial_xd_V_list[i], coeff_ind)
        return out

    def precomp_Vandermonde_hess_x_partial_xd(self, x, precomp=None):
        r""" Precompute Vandermonde matrices for the evaluation of :math:`\nabla^2_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`ndarray<ndarray>` [d,d]
            :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary with list of Vandermonde matrices for the computation
            of the gradient.
        """
        if precomp is None: precomp = {}
        try: hess_x_partial_xd_V_mat = precomp['hess_x_partial_xd_V_mat']
        except KeyError as e:
            try: V_list = precomp['V_list']
            except KeyError as e:
                self.precomp_evaluate(x, precomp)
                V_list = precomp['V_list']
            try: partial_x_V_list = precomp['partial_x_V_list']
            except KeyError as e:
                self.precomp_grad_x(x, precomp)
                partial_x_V_list = precomp['partial_x_V_list']
            try: partial2_x_V_list = precomp['partial2_x_V_list']
            except KeyError as e:
                self.precomp_hess_x(x, precomp)
                partial2_x_V_list = precomp['partial2_x_V_list']
            try: partial3_xd_V_last = precomp['partial3_xd_V_last']
            except KeyError as e:
                self.precomp_partial3_xd(x, precomp)
                partial3_xd_V_last = precomp['partial3_xd_V_last']
            hess_x_partial_xd_V_mat = np.empty((self.dim,self.dim),dtype=object)
            # TODO: Accelerate these loops?
            for d1 in range(self.dim):
                for d2 in range(self.dim):
                    hess_x_partial_xd_V = np.ones((x.shape[0], self.n_coeffs))
                    for i, midx in enumerate(self.multi_idxs):
                        for j, (idx, V1d,
                                pxV1d, p2xV1d) in enumerate(zip(midx, V_list,
                                                                partial_x_V_list,
                                                                partial2_x_V_list)):
                            if j == d1 == d2 == (self.dim-1):
                                hess_x_partial_xd_V[:,i] *= partial3_xd_V_last[:,idx]
                            elif j == d1 == d2 or j == d1 == (self.dim-1) \
                                 or j == d2 == (self.dim-1):
                                hess_x_partial_xd_V[:,i] *= p2xV1d[:,idx]
                            elif j == d1 or j == d2 or j == (self.dim-1):
                                hess_x_partial_xd_V[:,i] *= pxV1d[:,idx]
                            else:
                                hess_x_partial_xd_V[:,i] *= V1d[:,idx]
                    hess_x_partial_xd_V_mat[d1,d2] = hess_x_partial_xd_V
            precomp['hess_x_partial_xd_V_mat'] = hess_x_partial_xd_V_mat
        return precomp

    @counted
    def hess_x_partial_xd(self, x, precomp=None, idxs_slice=slice(None), *args, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) --
            :math:`\nabla^2_{\bf x}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        try:
            hess_x_partial_xd_V_mat = precomp['hess_x_partial_xd_V_mat']
        except (TypeError, KeyError) as e:
            try:
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
                partial2_x_V_list = precomp['partial2_x_V_list']
                partial3_xd_V_last = precomp['partial3_xd_V_last']
            except (TypeError, KeyError) as e:
                # Clean pre-existing
                precomp.pop('V_list', None)
                precomp.pop('partial_x_V_list', None)
                precomp.pop('partial2_x_V_list', None)
                precomp.pop('partial3_xd_V_last', None)
                # Ignoring slice
                idxs_slice = slice(None)
                # Compute
                precomp = self.precomp_evaluate(x, precomp)
                precomp = self.precomp_grad_x(x, precomp)
                precomp = self.precomp_hess_x(x, precomp)
                precomp = self.precomp_partial2_xd(x, precomp)
                precomp = self.precomp_partial3_xd(x, precomp)
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
                partial2_x_V_list = precomp['partial2_x_V_list']
                partial3_xd_V_last = precomp['partial3_xd_V_last']
            out = np.zeros( (x.shape[0], self.dim, self.dim) )
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate these loops?
            for d1 in range(self.dim):
                for d2 in range(self.dim):
                    for i, (c, midx) in enumerate(zip(self._coeffs,self.multi_idxs)):
                        tmp[:] = 1.
                        for j, (idx, V1d,
                                pxV1d, p2xV1d) in enumerate(zip(midx, V_list,
                                                                partial_x_V_list,
                                                                partial2_x_V_list)):
                            if j == d1 == d2 == (self.dim-1):
                                tmp *= partial3_xd_V_last[:,idx]
                            elif j == d1 == d2 or j == d1 == (self.dim-1) \
                                 or j == d2 == (self.dim-1):
                                tmp *= p2xV1d[:,idx]
                            elif j == d1 or j == d2 or j == (self.dim-1):
                                tmp *= pxV1d[:,idx]
                            else:
                                tmp *= V1d[:,idx]
                        out[:,d1,d2] += c * tmp
        else:
            out = np.zeros( (x.shape[0], self.dim, self.dim) )
            for i in range( self.dim ):
                for j in range(self.dim):
                    out[:,i,j] = np.dot( hess_x_partial_xd_V_mat[i,j], self._coeffs )
        return out

    @counted
    def grad_a_hess_x_partial_xd(self, x, precomp=None, idxs_slice=slice(None),
                                 *args, **kwars):
        r""" Evaluate :math:`\nabla_{\bf a}\nabla^2_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) --
            :math:`\nabla_{\bf a}\nabla^2_{\bf x}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        try:
            hess_x_partial_xd_V_mat = precomp['hess_x_partial_xd_V_mat']
        except (TypeError, KeyError) as e:
            try:
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
                partial2_x_V_list = precomp['partial2_x_V_list']
                partial3_xd_V_last = precomp['partial3_xd_V_last']
            except (TypeError, KeyError) as e:
                # Clean pre-existing
                precomp.pop('V_list', None)
                precomp.pop('partial_x_V_list', None)
                precomp.pop('partial2_x_V_list', None)
                precomp.pop('partial3_xd_V_last', None)
                # Ignoring slice
                idxs_slice = slice(None)
                # Compute
                precomp = self.precomp_evaluate(x, precomp)
                precomp = self.precomp_grad_x(x, precomp)
                precomp = self.precomp_hess_x(x, precomp)
                precomp = self.precomp_partial2_xd(x, precomp)
                precomp = self.precomp_partial3_xd(x, precomp)
                V_list = precomp['V_list']
                partial_x_V_list = precomp['partial_x_V_list']
                partial2_x_V_list = precomp['partial2_x_V_list']
                partial3_xd_V_last = precomp['partial3_xd_V_last']
            out = np.zeros( (x.shape[0], self.n_coeffs, self.dim, self.dim) )
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate these loops?
            for d1 in range(self.dim):
                for d2 in range(self.dim):
                    for i, (c, midx) in enumerate(zip(self._coeffs,self.multi_idxs)):
                        tmp[:] = 1.
                        for j, (idx, V1d,
                                pxV1d, p2xV1d) in enumerate(zip(midx, V_list,
                                                                partial_x_V_list,
                                                                partial2_x_V_list)):
                            if j == d1 == d2 == (self.dim-1):
                                tmp *= partial3_xd_V_last[:,idx]
                            elif j == d1 == d2 or j == d1 == (self.dim-1) \
                                 or j == d2 == (self.dim-1):
                                tmp *= p2xV1d[:,idx]
                            elif j == d1 or j == d2 or j == (self.dim-1):
                                tmp *= pxV1d[:,idx]
                            else:
                                tmp *= V1d[:,idx]
                        out[:,i,d1,d2] = tmp
        else:
            out = np.zeros( (x.shape[0], self.n_coeffs, self.dim, self.dim) )
            for i in range( self.dim ):
                for j in range(self.dim):
                    out[:,:,i,j] = hess_x_partial_xd_V_mat[i,j]
        return out

    def precomp_Vandermonde_partial2_xd(self, x, precomp=None):
        r""" Precompute multi-variate Vandermonde matrix for the evaluation of :math:`\partial^2_{x_d} f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary with Vandermonde matrix
        """
        if precomp is None: precomp = {}
        try: partial2_xd_V = precomp['partial2_xd_V']
        except KeyError as e:
            try: V_list = precomp['V_list']
            except (TypeError, KeyError) as e:
                self.precomp_evaluate(x, precomp)
                V_list = precomp['V_list']
            try: partial2_xd_V_last = precomp['partial2_xd_V_last']
            except (TypeError, KeyError) as e:
                self.precomp_partial2_xd(x, precomp)
                partial2_xd_V_last = precomp['partial2_xd_V_last']
            partial2_xd_V = np.ones((x.shape[0], self.n_coeffs))
            # TODO: Accelerate these loops?
            for i, midx in enumerate(self.multi_idxs):
                for idx, V1d in zip(midx[:-1], V_list[:-1]):
                    partial2_xd_V[:,i] *= V1d[:,idx]
                partial2_xd_V[:,i] *= partial2_xd_V_last[:,midx[-1]]
            precomp['partial2_xd_V'] = partial2_xd_V
        return precomp

    @counted
    def partial2_xd(self, x, precomp=None, idxs_slice=slice(None)):
        r""" Evaluate :math:`\partial^2_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial^2_{x_d} f_{\bf a}({\bf x})`
        """
        try:
            partial2_xd_V = precomp['partial2_xd_V']
        except (TypeError, KeyError) as e:
            try:
                V_list = precomp['V_list']
                partial2_xd_V_last = precomp['partial2_xd_V_last']
            except (TypeError, KeyError) as e:
                # Clean pre-existing
                precomp.pop('V_list', None)
                precomp.pop('partial2_xd_V_last', None)
                # Ignoring slice
                idxs_slice = slice(None)
                # Compute
                precomp = self.precomp_evaluate(x, precomp)
                precomp = self.precomp_partial2_xd(x, precomp)
                V_list = precomp['V_list']
                partial2_xd_V_last = precomp['partial2_xd_V_last']
            out = np.zeros(x.shape[0])
            tmp = np.ones(x.shape[0])
            # TODO: Accelerate these loops?
            for i, (c, midx) in enumerate(zip(self._coeffs,self.multi_idxs)):
                tmp[:] = 1.
                for idx, V1d in zip(midx[:-1], V_list[:-1]):
                    tmp *= V1d[:,idx]
                tmp *= partial2_xd_V_last[:,midx[-1]]
                out += c * tmp
        else:
            out = np.dot(partial2_xd_V, self._coeffs)
        return out

    @cached()
    @counted
    def grad_a_partial_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla_{\bf a}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            :math:`\nabla_{\bf a}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        try:
            partial_xd_V = precomp['partial_xd_V']
        except KeyError as e:
            if 'V_list' not in precomp or 'partial_xd_V_last' not in precomp:
                idxs_slice = slice(None)
                precomp.pop('V_list', None)
                precomp.pop('partial_xd_V_last', None)
            precomp = self.precomp_Vandermonde_partial_xd(x, precomp)
        except TypeError as e:
            idxs_slice = slice(None)
            precomp = self.precomp_Vandermonde_partial_xd(x, precomp)
        finally:
            partial_xd_V = precomp['partial_xd_V']
        return partial_xd_V[idxs_slice,:]

    @counted
    def hess_a_partial_xd(self, x, precomp=None, idxs_slice=slice(None), *args, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf a}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N,N`]) --
            :math:`\nabla^2_{\bf a}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        nc = self.n_coeffs
        return np.zeros( (x.shape[0], nc, nc) )
        
    def precomp_regression(self, x, precomp=None, *args, **kwargs):
        r""" Precompute necessary structures for the speed up of :func:`regression`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (dict): dictionary to be updated

        Returns:
           (:class:`dict<dict>`) -- dictionary of necessary strucutres
        """
        if precomp is None:
            precomp = {}
        precomp.update( self.precomp_evaluate(x) )
        return precomp

    def regression(self, f, fparams=None, d=None, qtype=None, qparams=None,
                   x=None, w=None, x0=None,
                   regularization=None, tol=1e-4, maxit=100,
                   batch_size=(None,None), mpi_pool=None, import_set=set()):
        r""" Compute :math:`{\bf a}^* = \arg\min_{\bf a} \Vert f - f_{\bf a} \Vert_{\pi}`.

        Args:
          f (:class:`Function` or :class:`ndarray<numpy.ndarray>` [:math:`m`]): function
            :math:`f` or its functions values
          d (Distribution): distribution :math:`\pi`
          fparams (dict): parameters for function :math:`f`
          qtype (int): quadrature type to be used for the approximation of
            :math:`\mathbb{E}_{\pi}`
          qparams (object): parameters necessary for the construction of the
            quadrature
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): quadrature points
            used for the approximation of :math:`\mathbb{E}_{\pi}`
          w (:class:`ndarray<numpy.ndarray>` [:math:`m`]): quadrature weights
            used for the approximation of :math:`\mathbb{E}_{\pi}`
          x0 (:class:`ndarray<numpy.ndarray>` [:math:`N`]): coefficients to be used
            as initial values for the optimization
          regularization (dict): defines the regularization to be used.
            If ``None``, no regularization is applied.
            If key ``type=='L2'`` then applies Tikonhov regularization with
            coefficient in key ``alpha``.
          tol (float): tolerance to be used to solve the regression problem.
          maxit (int): maximum number of iterations
          batch_size (:class:`list<list>` [2] of :class:`int<int>`): the list contains the
            size of the batch to be used for each iteration. A size ``1`` correspond
            to a completely non-vectorized evaluation. A size ``None`` correspond to a
            completely vectorized one.
          mpi_pool (:class:`mpi_map.MPI_Pool`): pool of processes to be used
          import_set (set): list of couples ``(module_name,as_field)`` to be imported
            as ``import module_name as as_field`` (for MPI purposes)

        Returns:
          (:class:`tuple<tuple>`(:class:`ndarray<numpy.ndarray>` [:math:`N`],
          :class:`list<list>`)) -- containing the :math:`N` coefficients and
          log information from the optimizer.

        .. seealso:: :func:`TransportMaps.TriangularTransportMap.regression`

        .. note:: the resulting coefficients :math:`{\bf a}` are automatically
           set at the end of the optimization. Use :func:`coeffs` in order
           to retrieve them.
        .. note:: The parameters ``(qtype,qparams)`` and ``(x,w)`` are mutually
          exclusive, but one pair of them is necessary.
        """
        self.mpi_pool = mpi_pool
        if (x is None) and (w is None):
            (x,w) = d.quadrature(qtype, qparams)
        params = {}
        params['x'] = x
        params['w'] = w
        params['regularization'] = regularization
        params['batch_size'] = batch_size
        params['nobj'] = 0
        params['nda_obj'] = 0
        options = {'maxiter': maxit,
                   'disp': False}
        # Zero initial condition
        if x0 is None:
            x0 = np.zeros( self.n_coeffs )
        # Precomputation
        if self.logger.getEffectiveLevel() <= logging.DEBUG:
            self.logger.debug("regression(): Precomputation started")
        # Prepare parameters
        if isinstance(f, np.ndarray):
            params['fvals'] = f
        else:
            scatter_tuple = (['x'], [x])
            bcast_tuple = (['precomp'], [fparams])
            params['fvals'] = mpi_map("evaluate", scatter_tuple=scatter_tuple,
                                      bcast_tuple=bcast_tuple,
                                      obj=f, mpi_pool=self.mpi_pool)
        # Init precomputation memory
        params['params1'] = {}
        mpi_alloc_dmem(params1=params['params1'], f1=self, mpi_pool=self.mpi_pool)
        
        scatter_tuple = (['x'], [x])
        mpi_map("precomp_regression", scatter_tuple=scatter_tuple,
                dmem_key_in_list=['params1'],
                dmem_arg_in_list=['precomp'],
                dmem_val_in_list=[params['params1']],
                obj='f1', obj_val=self,
                mpi_pool=self.mpi_pool,
                concatenate=False)
        if self.logger.getEffectiveLevel() <= logging.DEBUG:
            self.logger.debug("regression(): Precomputation ended")
        # Solve
        res = sciopt.minimize(self.regression_objective, x0, args=params, \
                              jac=self.regression_grad_a_objective,
                              method='BFGS', options=options, tol=tol)
        if not res['success']:
            self.logger.warn("Regression failure: " + res['message'])
        coeffs = res['x']
        self.coeffs = coeffs
        self.mpi_pool = None
        return (coeffs, res)

    @staticmethod
    def parse_xml_span_node(node, avars, totdim):
        from TransportMaps import XML_NAMESPACE
        span_node = node.find(XML_NAMESPACE + 'spanorder')
        # Retrieve the list of orders
        order_list = LinearSpanApproximation.order_list_from_xml_element(
            span_node, len(avars))
        # Retrieve the full list of basis
        full_basis_list = LinearSpanApproximation.basis_list_from_xml_element(
            node, totdim)
        # Retrieve the list of basis
        basis_list = [full_basis_list[v] for v in avars]
        return (span_node, order_list, full_basis_list, basis_list)

    @staticmethod
    def parse_xml_midx_node(node, avars, totdim):
        from TransportMaps import XML_NAMESPACE
        midxlist_node = node.find(XML_NAMESPACE + 'midxlist')
        # Retrieve the list of orders
        midx_list = LinearSpanApproximation.midx_list_from_xml_element(
            midxlist_node, len(avars))
        # Retrieve the full list of basis
        full_basis_list = LinearSpanApproximation.basis_list_from_xml_element(
            node, totdim)
        # Retrieve the list of basis
        basis_list = [full_basis_list[v] for v in avars]
        return (midxlist_node, midx_list, full_basis_list, basis_list)
        
    @staticmethod
    def from_xml_element(node, avars, totdim):
        # Check span type
        multidimtype = node.attrib['multidimtype'] 
        if multidimtype == 'tensorized':
            stype = node.attrib['spantype']
            if stype == 'total' or stype == 'full':
                span_node, order_list, full_basis_list, basis_list = \
                    LinearSpanApproximation.parse_xml_span_node(node, avars, totdim)
                return LinearSpanApproximation(
                    basis_list, spantype=stype, order_list=order_list,
                    full_basis_list=full_basis_list)
            elif stype == 'midx':
                midxlist_node, midx_list, full_basis_list, basis_list = \
                    LinearSpanApproximation.parse_xml_midx_node(node, avars, totdim)
                return LinearSpanApproximation(
                    basis_list, spantype=stype, midx_list=midx_list,
                    full_basis_list=full_basis_list)
            raise ValueError("No recognizable spantype provided (%s)" % stype)
        raise ValueError("No recognizable multidimtype provided (%s)" % multidimtype)

    @staticmethod
    def basis_list_from_xml_element(node, dim):
        import SpectralToolbox.Spectral1D as S1D
        import TransportMaps.XML as TMXML
        from TransportMaps import XML_NAMESPACE
        basis_list = [None for d in range(dim)]
        basis_nodes_list = node.findall(XML_NAMESPACE + 'spanbasis')
        for basis_node in basis_nodes_list:
            coord = basis_node.attrib['coord']
            rng = TMXML.vars_parser(coord, dim-1)
            base = S1D.from_xml_element(basis_node)
            for d in rng:
                basis_list[d] = base
        # Check all basis have been provided
        if None in basis_list:
            raise ValueError("The basis for some dimensions have not been defined")
        return basis_list

    @staticmethod
    def midx_list_from_xml_element(node, dim):
        from TransportMaps import XML_NAMESPACE
        import TransportMaps.XML as TMXML
        midx_list = [TMXML.midx_parser(midx_node.text) for midx_node in
                     node.findall(XML_NAMESPACE + 'midx') ]
        return midx_list

    @staticmethod
    def order_list_from_xml_element(node, dim):
        import TransportMaps.XML as TMXML
        from TransportMaps import XML_NAMESPACE
        order_list = [None for d in range(dim)]
        maxord_nodes_list = node.findall(XML_NAMESPACE + 'maxord')
        for maxord_node in maxord_nodes_list:
            coord = maxord_node.attrib['coord']
            rng = TMXML.vars_parser(coord, dim-1)
            order = int(maxord_node.text)
            for d in rng:
                order_list[d] = order
        # Check all orders have been provided
        if None in order_list:
            raise ValueError("The orders for some dimensions have not been defined")
        return order_list


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import warnings
import numpy as np
import numpy.linalg as npla
import scipy.stats as stats
import scipy.linalg as scila
import scipy.special as scispec

import SpectralToolbox.Spectral1D as S1D
import SpectralToolbox.SpectralND as SND

from TransportMaps.Misc import counted, cached
from TransportMaps.Distributions.DistributionBase import *
from TransportMaps.Distributions.TransportMapDistributions import *

__all__ = ['FrozenDistribution_1d',
           'GaussianDistribution', 'StandardNormalDistribution',
           'ChainGraphGaussianDistribution',
           'StarGraphGaussianDistribution',
           'GridGraphGaussianDistribution',
           'LogNormalDistribution', 'LogisticDistribution',
           'GammaDistribution', 'BetaDistribution',
           'WeibullDistribution', 'CauchyDistribution',
           'GumbelDistribution', 'BananaDistribution',
           'StudentTDistribution']

nax = np.newaxis

class FrozenDistribution_1d(Distribution):
    r""" [Abstract] Generic frozen distribution 1d
    """
    def __init__(self):
        super(FrozenDistribution_1d,self).__init__(1)
        self.base = StandardNormalDistribution(1)
        self.scipy_base = stats.norm()
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    def quadrature(self, qtype, qparams, mass=1, **kwargs):
        r""" Generate quadrature points and weights.

        Types of quadratures:

        Monte-Carlo (``qtype==0``)
           ``qparams``: (:class:`int`) -- number of samples

        Quasi-Monte-Carlo (``qtype==1``)
           ``qparams``: (:class:`int`) -- number of samples

        Latin-Hypercube-Sampling (``qtype==2``)
           ``qparams``: (:class:`int`) -- number of samples

        Gauss-quadrature (``qtype==3``)
           ``qparams``: (:class:`list<list>` [:math:`d`]) -- orders for
           each dimension

        .. seealso:: :func:`Distribution.quadrature`
        """
        if qtype == 0:
            # Monte Carlo sampling
            x = self.rvs(qparams)
            w = np.ones(qparams)/float(qparams)
        elif qtype == 1:
            # Quasi-Monte Carlo sampling
            raise NotImplementedError("Not implemented")
        elif qtype == 2:
            # Latin-Hyper cube sampling
            raise NotImplementedError("Todo")
        elif qtype == 3:
            # Gaussian quadrature
            (x1,w) = self.base.quadrature(3, qparams, mass=mass)
            x = self.dist.ppf( self.scipy_base.cdf(x1[:,0]) )[:,nax]
        else:
            raise ValueError("Quadrature type not recognized")
        return (x,w)

    @cached(caching=False)
    @counted
    def action_hess_x_log_pdf(self, x, dx, *args, **kwargs):
        return np.einsum('...ij,...j->...i',
                         self.hess_x_log_pdf(x, *args, **kwargs), dx )

#########################################################
# Definitions of Gaussian and Standard normal densities #
#########################################################
class GaussianDistribution(Distribution):
    r""" Multivariate Gaussian distribution :math:`\pi`

    Args:
      mu (:class:`ndarray<numpy.ndarray>` [:math:`d`]): mean vector
      sigma (:class:`ndarray<numpy.ndarray>` [:math:`d,d`]): covariance matrix
      precision (:class:`ndarray<numpy.ndarray>` [:math:`d,d`]): precision matrix
    """

    def __init__(self, mu, sigma=None, precision=None):
        if (sigma is not None) and (precision is not None):
            raise ValueError("The fields sigma and precision are mutually " +
                             "exclusive")
        super(GaussianDistribution,self).__init__(mu.shape[0])
        self._mu = None
        self._sigma = None
        self._precision = None
        self.mu = mu
        if sigma is not None:
            self.sigma = sigma
        if precision is not None:
            self.precision = precision

    @property
    def mu(self):
        return self._mu

    @mu.setter
    def mu(self, mu):
        if (self._sigma is not None and mu.shape[0] != self._sigma.shape[0]) or \
           (self._precision is not None and mu.shape[0] != self._precision.shape[0]):
            raise ValueError("Dimension d of mu and sigma/precision must be the same")
        self._mu = mu

    @property
    def sigma(self):
        return self._sigma

    @sigma.setter
    def sigma(self, sigma):
        if self.mu.shape[0] != sigma.shape[0] or self.mu.shape[0] != sigma.shape[1]:
            raise ValueError("Dimension d of mu and sigma must be the same")
        if self._sigma is None or np.any(self._sigma != sigma):
            self._sigma = sigma
            try:
                chol = scila.cho_factor(self._sigma, True) # True: lower triangular
            except scila.LinAlgError:
                # Obtain the square root from svd
                u,s,v = scila.svd(self._sigma)
                self.log_det_sigma = np.sum(np.log(s))
                self.det_sigma = np.exp( self.log_det_sigma )
                self.sampling_mat = u * np.sqrt(s)[np.newaxis,:]
                self.inv_sigma = np.dot(u * (1./s)[np.newaxis,:], v.T)
            else:
                self.det_sigma = np.prod(np.diag(chol[0]))**2.
                self.log_det_sigma = 2. * np.sum( np.log( np.diag(chol[0]) ) )
                self.sampling_mat = np.tril(chol[0])
                self.inv_sigma = scila.cho_solve(chol, np.eye(self.dim))

    @property
    def precision(self):
        return self._precision

    @precision.setter
    def precision(self, precision):
        if self.mu.shape[0] != precision.shape[0] or self.mu.shape[0] != precision.shape[1]:
            raise ValueError("Dimension d of mu and precision must be the same")
        if self._precision is None or np.any(self.inv_sigma != precision):
            self._precision = precision
            self.inv_sigma = precision
            chol = scila.cho_factor(self.inv_sigma, False) # False: upper triangular
            self.sigma = scila.cho_solve(chol, np.eye(self.dim))
            self.det_sigma = 1. / np.prod(np.diag(chol[0]))**2.
            self.log_det_sigma = - 2. * np.sum( np.log( np.diag(chol[0]) ) )
            self.sampling_mat = scila.solve_triangular(np.triu(chol[0]),
                                                       np.eye(self.dim),
                                                       lower=False)
            
    def rvs(self, m, *args, **kwargs):
        r""" Generate :math:`m` samples from the distribution.

        Args:
          m (int): number of samples

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- samples

        .. seealso:: :func:`Distribution.rvs`
        """
        z = stats.norm().rvs(m*self.dim).reshape((m,self.dim))
        samples = self._mu + np.dot( self.sampling_mat, z.T ).T
        return samples

    def quadrature(self, qtype, qparams, mass=1., **kwargs):
        r""" Generate quadrature points and weights.

        Types of quadratures:

        Monte-Carlo (``qtype==0``)
           ``qparams``: (:class:`int`) -- number of samples

        Quasi-Monte-Carlo (``qtype==1``)
           ``qparams``: (:class:`int`) -- number of samples

        Latin-Hypercube-Sampling (``qtype==2``)
           ``qparams``: (:class:`int`) -- number of samples

        Gauss-quadrature (``qtype==3``)
           ``qparams``: (:class:`list<list>` [:math:`d`]) -- orders for
           each dimension

        .. seealso:: :func:`Distribution.quadrature`
        """
        if qtype == 0:
            # Monte Carlo sampling
            x = self.rvs(qparams)
            w = np.ones(qparams)/float(qparams)
        elif qtype == 1:
            # Quasi-Monte Carlo sampling
            raise NotImplementedError("Not implemented")
        elif qtype == 2:
            # Latin-Hyper cube sampling
            raise NotImplementedError("Todo")
        elif qtype == 3:
            # Gaussian quadrature
            # Generate first a standard normal quadrature
            # then apply the Cholesky transform
            P = SND.PolyND( [S1D.HermiteProbabilistsPolynomial()] * self.dim )
            (x,w) = P.GaussQuadrature(qparams, norm=True)
            x = np.dot( self.sampling_mat, x.T ).T
            x += self._mu
            # For stability sort in ascending order of w
            srt_idxs = np.argsort(w)
            w = w[srt_idxs]
            x = x[srt_idxs,:]
        else:
            raise ValueError("Quadrature type not recognized")
        w *= mass
        return (x,w)

    @counted
    def pdf(self, x, *args, **kwargs):
        r""" Evaluate :math:`\pi(x)`

        .. seealso:: :func:`Distribution.pdf`
        """
        return np.exp( self.log_pdf(x) )

    @cached()
    @counted
    def log_pdf(self, x, *args, **kwargs):
        r""" Evaluate :math:`\log\pi(x)`

        .. seealso:: :func:`Distribution.log_pdf`
        """
        b = x - self._mu
        sol = np.dot( self.inv_sigma, b.T ).T
        out = - .5 * np.einsum('...i,...i->...', b, sol) \
              - self.dim * .5 * np.log(2.*np.pi) \
              - .5 * self.log_det_sigma
        return out.flatten()

    @cached()
    @counted
    def grad_x_log_pdf(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x}\log\pi(x)`

        .. seealso:: :func:`Distribution.grad_x_log_pdf`
        """
        b = x - self._mu
        return - np.dot( self.inv_sigma, b.T ).T

    @counted
    def hess_x_log_pdf(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf x}\log\pi(x)`

        .. seealso:: :func:`Distribution.hess_x_log_pdf`
        """
        return - np.ones(x.shape[0])[:,nax,nax] * self.inv_sigma[nax,:,:]

    @counted
    def action_hess_x_log_pdf(self, x, dx, *args, **kwargs):
        r""" Evaluate :math:`\langle \nabla^2_{\bf x} \log \pi({\bf x}), \delta{\bf x}\rangle`

        .. seealso:: :func:`Distribution.action_hess_x_log_pdf`
        """
        return - np.dot( dx, self.inv_sigma )

    def mean_log_pdf(self):
        r""" Evaluate :math:`\mathbb{E}_{\pi}[\log \pi]`.

        .. seealso:: :func:`Distribution.mean_log_pdf`
        """
        return - .5 * ( self.dim * np.log(2*np.pi) + self.dim + self.log_det_sigma )

class StandardNormalDistribution(GaussianDistribution, ProductDistribution):
    r""" Multivariate Standard Normal distribution :math:`\pi`.

    Args:
      d (int): dimension
    """

    def __init__(self, dim):
        self._mu = np.zeros(dim)
        self._sigma = np.eye(dim)
        self._precision = np.eye(dim)
        self.log_det_sigma = 0.
        self.det_sigma = 1.
        self.sampling_mat = np.eye(dim)
        self.inv_sigma = np.eye(dim)
        Distribution.__init__(self, dim)

    def rvs(self, m, *args, **kwargs):
        r""" Generate :math:`m` samples from the distribution.

        Args:
          m (int): number of samples

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) -- samples

        .. seealso:: :func:`Distribution.rvs`
        """
        return stats.norm().rvs(m*self.dim).reshape((m,self.dim))

    def get_component(self, avars):
        r""" Return the measure :math:`\nu_{a_1}\times\cdots\times\nu_{a_k} = \mathcal{N}(0,{\bf I}_k)`

        Args:
          avars (list): list of coordinates to extract from :math:`\nu`
        """
        return StandardNormalDistribution(len(avars))

class ChainGraphGaussianDistribution(GaussianDistribution):
    def __init__(self, dim, edge_strength=.45):
        self.dim = dim
        self.edge_strength = edge_strength
        mu = np.zeros(dim)
        precision = self.omega()
        super(ChainGraphGaussianDistribution, self).__init__(mu, precision=precision)

    def omega(self):
        # create tridiagonal matrix 
        omega = np.eye(self.dim) + self.edge_strength*np.eye(self.dim,k=-1) + self.edge_strength*np.eye(self.dim,k=+1)
        # compute unnormalized sigma (covariance matrix)
        sigma_temp = np.linalg.inv(omega)
        # extract standard deviations from sigma_temp
        std_temp = np.diag(np.sqrt(np.diag(sigma_temp)))
        # scale omega by std_temp
        omega = np.dot(np.dot(std_temp, omega), std_temp)
        return omega

    def sigma(self):
        # return inverse of omega
        sigma = np.linalg.inv(self.omega)
        return sigma

    def rvs(self, m):
        # compute K, where K K^T = sigma
        K = np.linalg.cholesky(self.sigma)
        # generate 'base' samples from standard normal
        x = stats.norm().rvs(m*self.dim).reshape((m,self.dim))
        # return m realizations
        samples = np.dot(K, x.T).T
        return samples

    @property
    def nonzero_idxs(self):
        # find zero elements in omega to determine active_vars
        omegaLower = np.tril(self.omega())
        active_vars = []
        for i in range(self.dim):
            actives = np.where(omegaLower[i,:] != 0)
            active_list = list(set(actives[0]) | set([i]))
            active_list.sort(key=int)
            active_vars.append(active_list)
        return active_vars

    @property
    def graph(self):
        graph = np.zeros([self.dim,self.dim])
        graph[self.omega() != 0] = 1
        return graph

    @property
    def n_edges(self):
        return self.dim-1

class StarGraphGaussianDistribution(GaussianDistribution):
    def __init__(self, dim, edge_strength=.45):
        self.dim = dim
        self.edge_strength = edge_strength
        mu = np.zeros(dim)
        precision = self.omega()
        super(StarGraphGaussianDistribution, self).__init__(mu, precision=precision)

    def omega(self):
        # create star matrix 
        omega = np.eye(self.dim)
        omega[0,0] = self.dim
        omega[0,1:] = self.edge_strength
        omega[1:,0] = self.edge_strength
        # compute unnormalized sigma (covariance matrix)
        sigma_temp = np.linalg.inv(omega)
        # extract standard deviations from sigma_temp
        std_temp = np.diag(np.sqrt(np.diag(sigma_temp)))
        # scale omega by std_temp
        omega = np.dot(np.dot(std_temp, omega), std_temp)
        return omega

    def sigma(self):
        # return inverse of omega
        sigma = np.linalg.inv(self.omega)
        return sigma

    def rvs(self, m):
        # compute K, where K K^T = sigma
        K = np.linalg.cholesky(self.sigma)
        # generate 'base' samples from standard normal
        x = stats.norm().rvs(m*self.dim).reshape((m,self.dim))
        # return m realizations
        samples = np.dot(K, x.T).T
        return samples

    @property
    def nonzero_idxs(self):
        # find zero elements in omega to determine active_vars
        omegaLower = np.tril(self.omega())
        active_vars = []
        for i in range(self.dim):
            actives = np.where(omegaLower[i,:] != 0)
            active_list = list(set(actives[0]) | set([i]))
            active_list.sort(key=int)
            active_vars.append(active_list)
        return active_vars

    @property
    def graph(self):
        graph = np.zeros([self.dim,self.dim])
        graph[self.omega() != 0] = 1
        return graph

    @property
    def n_edges(self):
        return self.dim-1

class GridGraphGaussianDistribution(GaussianDistribution):
    def __init__(self, dim, edge_strength= 1.0):
        if (np.sqrt(dim) - int(np.sqrt(dim))) != 0:
            raise ValueError('Input dimension must be a square number')
        self.dim = dim
        self.edge_strength = edge_strength
        mu = np.zeros(dim)
        precision = self.omega()
        super(GridGraphGaussianDistribution, self).__init__(mu, precision=precision)

    def omega(self):
        dim_sq_root = int(np.sqrt(self.dim))

        # declare grid coordinates
        coords = self.zigzag(dim_sq_root)
        n_coords = len(coords)
    
        # create zero matrix
        omega = np.zeros((n_coords, n_coords))

        # pull out all coordinates
        all_coords = list(coords.values())

        # add all edges for the grid graph
        for i in range(n_coords):
            coord_val = coords[i];
            new_coords = [(coord_val[0],coord_val[1]+1), 
                          (coord_val[0],coord_val[1]-1), 
                          (coord_val[0]+1,coord_val[1]),
                          (coord_val[0]-1,coord_val[1])]
            for j in range(len(new_coords)):
                if new_coords[j] in all_coords:
                    coord_idx = all_coords.index(new_coords[j])
                    omega[i, coord_idx] = self.edge_strength
                    omega[coord_idx, i] = self.edge_strength

        # set the diagonal appropriately
        max_val = np.ceil(np.max(np.sum(np.abs(omega), axis = 0))) + 1
        omega = omega + max_val*np.eye(n_coords)
        
        # compute unnormalized sigma (covariance matrix)
        sigma_temp = np.linalg.inv(omega)
        # extract standard deviations from sigma_temp
        std_temp = np.diag(np.sqrt(np.diag(sigma_temp)))
        # scale omega by std_temp
        omega = np.dot(np.dot(std_temp, omega), std_temp)
        return omega

    def zigzag(self,n):
        # zig-zag pattern returns bijection between graph coordinates and ordering
        indexorder = sorted(((x,y) for x in range(n) for y in range(n)),
                    key = lambda p: (p[0]+p[1], -p[1] if (p[0]+p[1]) % 2 else p[1]) )
        return dict((n,index) for n,index in enumerate(indexorder))

    def sigma(self):
        # return inverse of omega
        sigma = np.linalg.inv(self.omega)
        return sigma

    def rvs(self, m):
        # compute K, where K K^T = sigma
        K = np.linalg.cholesky(self.sigma)
        # generate 'base' samples from standard normal
        x = stats.norm().rvs(m*self.dim).reshape((m,self.dim))
        # return m realizations
        samples = np.dot(K, x.T).T
        return samples

    @property
    def nonzero_idxs(self):
        #dim_sq = np.power(self.dim,2)

        # extract lower triangular matrix
        omegaLower = np.tril(self.omega())

        # add edges by...
        # variable elimination moving from highest node (dim-1) to node 2 (at most)
        for i in range(self.dim-1,1,-1):
            non_zero_ind  = np.where(omegaLower[i,:i] != 0)[0]
            if len(non_zero_ind) > 1:
                co_parents = list(itertools.combinations(non_zero_ind,2))
                for j in range(len(co_parents)):
                    row_index = max(co_parents[j])
                    col_index = min(co_parents[j])
                    omegaLower[row_index, col_index] = 1.0

        # find zero elements in chordal omega to determine active_vars
        active_vars = []
        for i in range(self.dim):
            actives = np.where(omegaLower[i,:] != 0)
            active_list = list(set(actives[0]) | set([i]))
            active_list.sort(key=int)
            active_vars.append(active_list)

        return active_vars

    @property
    def graph(self):
        graph = np.zeros([self.dim,self.dim])
        graph[self.omega() != 0] = 1
        return graph

    @property
    def n_edges(self):
        return 2.*(np.sqrt(self.dim) - 1)**2

###############################################################
# Definition of miscellaneous densities (No sampling defined) #
###############################################################
class LogNormalDistribution(FrozenDistribution_1d):
    def __init__(self, s, mu, scale):
        super(LogNormalDistribution,self).__init__()
        self.s = s
        self.mu = mu
        self.scale = scale
        self.dist = stats.lognorm(s=s,
                                  loc=mu,
                                  scale=scale)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    @counted
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf( x ).flatten()
    @counted
    def grad_x_pdf(self, x, *args, **kwargs):
        s = self.s
        m = self.mu
        d = self.dist
        return - d.pdf(x) * ( 1./(x-m) + np.log(x-m)/(s**2.*(x-m)) )
    @counted
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf( x ).flatten()
    @counted
    def grad_x_log_pdf(self, x, *args, **kwargs):
        s = self.s
        m = self.mu
        sc = self.scale
        return - 1./(x-m) * (np.log((x-m)/sc)/s**2. + 1)
    @counted
    def hess_x_log_pdf(self, x, *args, **kwargs):
        s = self.s
        m = self.mu
        sc = self.scale
        return (1./(x-m)**2. * ( (np.log((x-m)/sc) + s**2. - 1.)/s**2. ))[:,:,nax]

class LogisticDistribution(FrozenDistribution_1d):
    def __init__(self, mu, s):
        super(LogisticDistribution,self).__init__()
        self.mu = mu
        self.s = s
        self.dist = stats.logistic(loc=mu,scale=s)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    @counted
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    @counted
    def log_pdf(self, x, *args, **kwargs):
        # Log pdf with modified asymptotic behavior
        out = np.zeros(x.shape)
        g20 = (x >= -20)
        l20 = (x < -20)
        out[g20] = self.dist.logpdf(x[g20]).flatten()
        out[l20] = (x[l20].flatten() - self.mu)/self.s
        return out.flatten()
    @counted
    def grad_x_log_pdf(self, x, *args, **kwargs):
        mu = self.mu
        s = self.s
        out = np.zeros(x.shape)
        g20 = (x >= -20)
        l20 = (x < -20)
        g = np.exp(-(x[g20]-mu)/s)
        out[g20] = -1./s + 2./s * g/(1+g)
        out[l20] = 1./s
        return out
    @counted
    def hess_x_log_pdf(self, x, *args, **kwargs):
        mu = self.mu
        s = self.s
        out = np.zeros(x.shape)
        g20 = (x >= -20)
        l20 = (x < -20)
        g = np.exp(-(x[g20]-mu)/s)
        out[g20] = (- 2./s**2. * g/(1+g)**2.)
        out[l20] = 0.
        return out[:,:,nax]
    # def nabla3_x_log_pdf(self, x, params=None):
    #     mu = self.mu
    #     s = self.s
    #     g = np.exp(-(x-mu)/s)
    #     return (2./s**3. * g*(1-g)/(1+g)**3.)[:,:,nax,nax]

class GammaDistribution(FrozenDistribution_1d):
    def __init__(self, kappa, theta):
        super(GammaDistribution,self).__init__()
        self.kappa = kappa
        self.theta = theta
        self.dist = stats.gamma(kappa, scale=theta)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    @counted
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    @counted
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf(x).flatten()
    @counted
    def grad_x_log_pdf(self, x, *args, **kwargs):
        k = self.kappa
        t = self.theta
        return (k-1.)/x - 1/t
    @counted
    def hess_x_log_pdf(self, x, *args, **kwargs):
        k = self.kappa
        return ((1.-k)/x**2.)[:,:,nax]
    @counted
    def nabla3_x_log_pdf(self, x, *args, **kwargs):
        k = self.kappa
        return (2.*(k-1.)/x**3.)[:,:,nax,nax]

class BetaDistribution(FrozenDistribution_1d):
    def __init__(self, alpha, beta):
        super(BetaDistribution,self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.dist = stats.beta(alpha, beta)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    @counted
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    @counted
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf(x).flatten()
    @counted
    def grad_x_log_pdf(self, x, *args, **kwargs):
        a = self.alpha
        b = self.beta
        return (a-1.)/x + (b-1.)/(x-1.)
    @counted
    def hess_x_log_pdf(self, x, *args, **kwargs):
        a = self.alpha
        b = self.beta
        out = (1.-a)/x**2. + (1-b)/(x-1.)**2.
        return out[:,:,nax]
    @counted
    def nabla3_x_log_pdf(self, x, *args, **kwargs):
        a = self.alpha
        b = self.beta
        out = 2.*(a-1.)/x**3. + 2.*(b-1.)/(x-1.)**3.
        return out[:,:,nax,nax]

class GumbelDistribution(FrozenDistribution_1d):
    def __init__(self, mu, beta):
        super(GumbelDistribution,self).__init__()
        self.mu = mu
        self.beta = beta
        self.dist = stats.gumbel_r(loc=mu, scale=beta)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    @counted
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    @counted
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf(x).flatten()
    @counted
    def grad_x_log_pdf(self, x, *args, **kwargs):
        m = self.mu
        b = self.beta
        z = (x-m)/b
        return (np.exp(-z)-1.)/b
    @counted
    def hess_x_log_pdf(self, x, *args, **kwargs):
        m = self.mu
        b = self.beta
        z = (x-m)/b
        return (- np.exp(-z)/b**2.)[:,:,nax]
    @counted
    def nabla3_x_log_pdf(self, x, *args, **kwargs):
        m = self.mu
        b = self.beta
        z = (x-m)/b
        return (np.exp(-z)/b**3.)[:,:,nax,nax]

class WeibullDistribution(FrozenDistribution_1d):
    def __init__(self, c, mu=0., sigma=1.):
        super(WeibullDistribution,self).__init__()
        self.c = c
        self.mu = mu
        self.sigma=sigma
        self.dist = stats.weibull_min(c=self.c, loc=self.mu, scale=self.sigma)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    @counted
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    @counted
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf(x).flatten()
    @counted
    def grad_x_log_pdf(self, x, *args, **kwargs):
        c = self.c
        m = self.mu
        s = self.sigma
        out = (c-1.)/(x-m) - c/s * ((x-m)/s)**(c-1.)
        return out
    @counted
    def hess_x_log_pdf(self, x, *args, **kwargs):
        c = self.c
        m = self.mu
        s = self.sigma
        out = - (c-1.)/(x-m)**2. - (c*(c-1.))/s**2. * ((x-m)/s)**(c-2.)
        return out[:,:,nax]

class CauchyDistribution(FrozenDistribution_1d):
  def __init__(self, loc, scale):
    self.loc = loc
    self.scale = scale
  def pdf(self,x,params=None):
    out = stats.cauchy.pdf(x, loc = self.loc, scale = self.scale)
    return out.flatten()
  def log_pdf(self,x,params=None):
    out = stats.cauchy.logpdf(x, loc = self.loc, scale = self.scale)
    return out.flatten()
  def rvs(self,n):
    out = stats.cauchy.rvs(size=n)
    return out
  def quadrature(self,qtype,qparams):
    if qtype == 0:
      x = qparams
      l = qparams.shape[0]
      w = np.ones(l)/l
    else:
      raise NotImplementedError("Not implemented")
    return (x,w)

class StudentTDistribution(FrozenDistribution_1d):
    def __init__(self, df, mu=0., sigma=1.):
        super(StudentTDistribution,self).__init__()
        if df < 1:
            raise AttributeError("df must be >= 1")
        self.mu = mu
        self.sigma = sigma
        self.df = df
        self.dist = stats.t(df, loc=mu, scale=sigma)
    def rvs(self, n, *args, **kwargs):
        return self.dist.rvs(n).reshape((n,1))
    @counted
    def pdf(self, x, *args, **kwargs):
        return self.dist.pdf(x).flatten()
    @counted
    def log_pdf(self, x, *args, **kwargs):
        return self.dist.logpdf(x).flatten()
    @counted
    def grad_x_log_pdf(self, x, *args, **kwargs):
        m = self.mu
        s = self.sigma
        k = self.df
        return - (k+1)*(x-m)/(k*s**2 + (x-m)**2)
    @counted
    def hess_x_log_pdf(self, x, *args, **kwargs):
        m = self.mu
        s = self.sigma
        k = self.df
        out =  - (k+1)/(k*s**2 + (x-m)**2) + \
               2*(k+1)*(x-m)**2/(k*s**2 + (x-m)**2)**2
        return out[:,:,nax]
        
class BananaDistribution(PushForwardTransportMapDistribution):
    def __init__(self, a, b, mu, sigma2):
        import TransportMaps.Maps as MAPS
        gauss_map = MAPS.LinearTransportMap( mu, npla.cholesky(sigma2) )
        ban_map = MAPS.FrozenBananaMap(a, b)
        tm = MAPS.CompositeMap(ban_map, gauss_map)
        base_distribution = StandardNormalDistribution(2)
        super(BananaDistribution, self).__init__(tm, base_distribution)

class RelaxedRademacherDistribution(Distribution):
    def __init__(self, dim):
        if dim%2 != 0:
            raise ValueError("Input dimension must be an even number")
        self.dim = dim

    def rvs(self, m):
        rvs = np.zeros([m, self.dim])
        for i in range(0,self.dim,2):
            rvs[:,i] = stats.norm.rvs(size=m)
            unif = 2*stats.uniform.rvs(size=m) - 1
            rvs[:,i+1] = rvs[:,i] * unif
        return rvs

    @property
    def graph(self):
        graph = np.eye(self.dim)
        for i in range(0,self.dim,2):
            graph[i,i+1] = 1
            graph[i+1,i] = 1
        return graph

    @property
    def nonzero_idxs(self):
        # find zero elements in omega to determine active_vars
        graphLower = np.tril(self.graph())
        active_vars = []
        for i in range(self.dim):
            actives = np.where(graphLower[i,:] != 0)
            active_list = list(set(actives[0]) | set([i]))
            active_list.sort(key=int)
            active_vars.append(active_list)
        return active_vars
        
    @property
    def n_edges(self):
        return self.dim/2.

class ButterflyDistribution(Distribution):
    def __init__(self, dim):
        if dim%2 != 0:
            raise ValueError("Input dimension must be an even number")
        self.dim = dim

    def rvs(self, m):
        rvs = np.zeros([m, self.dim])
        for i in range(0,self.dim,2):
            rvs[:,i] = stats.norm.rvs(size=m)
            norm = stats.norm.rvs(size=m)
            rvs[:,i+1] = rvs[:,i] * norm
        return rvs

    @property
    def graph(self):
        graph = np.eye(self.dim)
        for i in range(0,self.dim,2):
            graph[i,i+1] = 1
            graph[i+1,i] = 1
        return graph

    @property
    def nonzero_idxs(self):
        # find zero elements in omega to determine active_vars
        graphLower = np.tril(self.graph())
        active_vars = []
        for i in range(self.dim):
            actives = np.where(graphLower[i,:] != 0)
            active_list = list(set(actives[0]) | set([i]))
            active_list.sort(key=int)
            active_vars.append(active_list)
        return active_vars
        
    @property
    def n_edges(self):
        return self.dim/2.



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import logging
import numpy as np
import numpy.linalg as npla

from TransportMaps.Misc import cached, counted, get_sub_cache
from .ParametricFunctionApproximationBase import \
    ParametricFunctionApproximation
from TransportMaps.Distributions.DistributionBase import Distribution
from TransportMaps.Distributions.FrozenDistributions import StandardNormalDistribution as snd 

__all__ = ['ProductDistributionParametricPullbackComponentFunction']

nax = np.newaxis

class ProductDistributionParametricPullbackComponentFunction(
        ParametricFunctionApproximation):
    r""" Parametric function :math:`f[{\bf a}](x_{1:k}) = \log\pi\circ T_k[{\bf a}](x_{1:k}) + \log\partial_{x_k}T_k[{\bf a}](x_{1:k})`

    Args:
      transport_map_component (MonotonicFunctionApproximation): component :math:`T_k`
        of monotone map :math:`T`
      base_distribution (Distribution): distribution :math:`\pi`
    """
    def __init__(self, transport_map_component, base_distribution):
        if base_distribution.dim != 1:
            raise AttributeError("The dimension of the distribution must be 1")
        super(ParametricFunctionApproximation, self).__init__(transport_map_component.dim)
        self.tmap_component = transport_map_component
        self.base_distribution = base_distribution

    @property
    def coeffs(self):
        r""" Get the coefficients :math:`{\bf a}` of the function

        .. seealso:: :func:`ParametricFunctionApproximation.coeffs`
        """
        return self.tmap_component.coeffs

    @property
    def n_coeffs(self):
        r""" Get the number :math:`N` of coefficients
        
        .. seealso:: :func:`ParametricFunctionApproximation.n_coeffs`
        """
        return self.tmap_component.n_coeffs

    @coeffs.setter
    def coeffs(self, coeffs):
        r""" Set the coefficients :math:`{\bf a}` of the distribution

        .. seealso:: :func:`ParametricFunctionApproximation.coeffs`
        """
        self.tmap_component.coeffs = coeffs

    @cached([('pi',None), ('ti',None)])
    @counted
    def evaluate(self, x, params={}, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`f[{\bf a}](x_{1:k})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,k`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (dict): cached values

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- evaluations
        """
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        params_pi = params.get('params_pi')
        params_ti = params.get('params_t')
        # Retrieve cache
        pi_cache, ti_cache = get_sub_cache(cache, ('pi',None), ('ti',None))
        # Evaluate
        ev = self.tmap_component.evaluate(x, params_ti, idxs_slice=idxs_slice,
                                          cache=ti_cache)
        lpdf = self.base_distribution.log_pdf(ev[:,nax], params_pi, cache=pi_cache)
        lgxd = np.log( self.tmap_component.partial_xd(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache) )
        return lpdf + lgxd

    @cached([('pi',None), ('ti',None)])
    @counted
    def grad_a(self, x, params={}, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla_{\bf a}f[{\bf a}](x_{1:k})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,k`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (dict): cached values

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,n`]) -- evaluations
        """
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        params_pi = params['params_pi']
        params_ti = params['params_t']
        # Retrieve cache
        pi_cache, ti_cache = get_sub_cache(cache, ('pi',None), ('ti',None))
        # Evaluate
        ev = self.tmap_component.evaluate(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        gxlpdf = self.base_distribution.grad_x_log_pdf(
            ev[:,nax], params_pi, cache=pi_cache)[:,0]
        ga = self.tmap_component.grad_a(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        gxd = self.tmap_component.partial_xd(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        gagxd = self.tmap_component.grad_a_partial_xd(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        out = gxlpdf[:,nax] * ga + gagxd / gxd[:,nax]
        return out

    @cached([('pi',None), ('ti',None)], False)
    @counted
    def hess_a(self, x, params={}, idxs_slice=slice(None), cache=None):
        r""" Evaluate :math:`\nabla^2_{\bf a}f[{\bf a}](x_{1:k})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,k`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (dict): cached values

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,n`]) -- evaluations
        """
        if x.shape[1] != self.dim:
            raise ValueError("dimension mismatch")
        params_pi = params.get('params_pi')
        params_ti = params.get('params_t')
        # Retrieve cache
        pi_cache, ti_cache = get_sub_cache(cache, ('pi',None), ('ti',None))
        # Evaluate
        # First term
        ev = self.tmap_component.evaluate(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        hxlpdf = self.base_distribution.hess_x_log_pdf(
            ev[:,nax], params_pi, cache=pi_cache)[:,0,0]
        ga = self.tmap_component.grad_a(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        t1 = ga[:,nax,:] * ga[:,:,nax]
        t1 *= hxlpdf[:,nax,nax]
        # Second term
        gxlpdf = self.base_distribution.grad_x_log_pdf(
            ev[:,nax], params_pi, cache=pi_cache)[:,0]
        ha = self.tmap_component.hess_a(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        t2 = gxlpdf[:,nax,nax] * ha
        # Third term
        gxd = self.tmap_component.partial_xd(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        hagxd = self.tmap_component.hess_a_partial_xd(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        t3 = hagxd / gxd[:,nax,nax]
        # Fourth term
        gagxd = self.tmap_component.grad_a_partial_xd(
            x, params_ti, idxs_slice=idxs_slice, cache=ti_cache)
        t4 = (gagxd[:,nax,:] * gagxd[:,:,nax]) / gxd[:,nax,nax]**2
        # Output
        out = t1 + t2 + t3 - t4
        return out

    def grad_x(self, x, params={}, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla_{\bf x} \log T_{k}^\sharp \pi({\bf x_{1:k}})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,k`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- values of
            :math:`\nabla_{\bf x} \log T_{k}^\sharp \pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_ti = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_ti = None
        ev = self.tmap_component.evaluate(x, params_ti, idxs_slice)
        gxlpdf = self.base_distribution.grad_x_log_pdf(ev[:,nax], params_pi)[:,0]
        gx = self.tmap_component.grad_x(x, params_ti, idxs_slice)
        gxd = self.tmap_component.partial_xd(x, params_ti, idxs_slice)
        gxgxd = self.tmap_component.grad_x_partial_xd(x, params_ti, idxs_slice)
        out = gxlpdf[:,nax] * gx + gxgxd / gxd[:,nax]
        return out

    def hess_x(self, x, params={}, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla^2_{\bf x} \log T_{k}^\sharp \pi({\bf x_{1:k}})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,k`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,k,k`]) -- values of
            :math:`\nabla^2_{\bf x} \log T_{k}^\sharp \pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_ti = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_ti = None
        # First term
        ev = self.tmap_component.evaluate(x, params_ti, idxs_slice)
        hxlpdf = self.base_distribution.hess_x_log_pdf(ev[:,nax], params_pi)[:,0,0]
        gx = self.tmap_component.grad_x(x, params_ti, idxs_slice)
        t1 = gx[:,nax,:] * gx[:,:,nax]
        t1 *= hxlpdf[:,nax,nax]
        # Second term
        gxlpdf = self.base_distribution.grad_x_log_pdf(ev[:,nax], params_pi)[:,0]
        hx = self.tmap_component.hess_x(x, params_ti, idxs_slice)
        t2 = gxlpdf[:,nax,nax] * hx
        # Third term
        gxd = self.tmap_component.partial_xd(x, params_ti, idxs_slice)
        hxgxd = self.tmap_component.hess_x_partial_xd(x, params_ti, idxs_slice)
        t3 = hxgxd / gxd[:,nax,nax]
        # Fourth term
        gxgxd = self.tmap_component.grad_x_partial_xd(x, params_ti, idxs_slice)
        t4 = (gxgxd[:,nax,:] * gxgxd[:,:,nax]) / gxd[:,nax,nax]**2
        # Output
        out = t1 + t2 + t3 - t4
        return out

    def grad_a_hess_x(self, x, params={}, idxs_slice=slice(None)):
        r""" Evaluate :math:`\nabla_{\bf a} \nabla^2_{\bf x} \log T_{k}^\sharp \pi({\bf x_{1:k}})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,k`]): evaluation points
          params (dict): parameters with keys ``params_pi``, ``params_t``
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,n,k,k`]) -- values of
            :math:`\nabla^2_{\bf x} \log T_{k}^\sharp \pi` at the ``x`` points.
        """
        try:
            params_pi = params['params_pi']
        except (KeyError,TypeError):
            params_pi = None
        try:
            params_ti = params['params_t']
        except (KeyError,TypeError):
            idxs_slice = slice(None)
            params_ti = None

        # check if base distribution is standard normal
        if issubclass(type(self.base_distribution), snd) == False:
            raise NotImplementedError('Base distribution must be standard normal')

        # First term
        ev = self.tmap_component.evaluate(x, params_ti, idxs_slice)
        hxlpdf = self.base_distribution.hess_x_log_pdf(ev[:,nax], params_pi)[:,0,0]
        gx = self.tmap_component.grad_x(x, params_ti, idxs_slice)
        gagx = self.tmap_component.grad_a_grad_x(x, params_ti, idxs_slice)
        t1 = hxlpdf[:,nax,nax,nax]*gagx[:,:,:,nax]*gx[:,nax,nax,:]

        # Second term
        t2 = hxlpdf[:,nax,nax,nax]*gagx[:,:,nax,:]*gx[:,nax,:,nax]
        
        # Third term
        ga = self.tmap_component.grad_a(x, params_ti, idxs_slice)
        hx = self.tmap_component.hess_x(x, params_ti, idxs_slice)
        t3 = hxlpdf[:,nax,nax,nax]*ga[:,:,nax,nax]*hx[:,nax,:,:]

        # Fourth term
        gxlpdf = self.base_distribution.grad_x_log_pdf(ev[:,nax], params_pi)[:,0]
        gahx = self.tmap_component.grad_a_hess_x(x, params_ti, idxs_slice)
        t4 = gxlpdf[:,nax,nax,nax]*gahx

        # Fifth term (log det term)
        # Compute grad_a_hess_x_sum
        a = self.tmap_component
        dxk = a.partial_xd(x)
        dadx2dxk = a.grad_a_hess_x_partial_xd(x)
        dadxk    = a.grad_a_partial_xd(x)
        dadxdxk  = a.grad_a_grad_x_partial_xd(x)
        dx2dxk   = a.hess_x_partial_xd(x)
        dxdxkT   = a.grad_x_partial_xd(x)
        dxdxkT2  = dxdxkT[:,nax,:,nax] * dxdxkT[:,nax,nax,:]
        B = dadxdxk[:,:,:,nax]*dxdxkT[:,nax,nax,:]
        t5 = (dadx2dxk / dxk[:,nax,nax,nax]) - \
                (dx2dxk[:,nax,:,:]*dadxk[:,:,nax,nax])/(dxk**2.)[:,nax,nax,nax] - \
                (B + B.transpose((0,1,3,2)))/(dxk**2.)[:,nax,nax,nax] + \
                2*(dxdxkT2*dadxk[:,:,nax,nax])/(dxk**3.)[:,nax,nax,nax]

        # Output 
        out = t1 + t2 + t3 + t4 + t5
        return out



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import warnings
import scipy.linalg as scialg
from scipy.optimize import OptimizeResult
from scipy.optimize.linesearch import line_search_wolfe1, line_search_wolfe2, LineSearchWarning
import numpy as np

__all__ = ['minimize_newtonlu','minimize_newtoncglu']

_epsilon = np.sqrt(np.finfo(float).eps)

# standard status messages of optimizers
_status_message = {'success': 'Optimization terminated successfully.',
                   'maxfev': 'Maximum number of function evaluations has '
                              'been exceeded.',
                   'maxiter': 'Maximum number of iterations has been '
                              'exceeded.',
                   'pr_loss': 'Desired error not necessarily achieved due '
                              'to precision loss.'}

class _LineSearchError(RuntimeError):
    pass

def _line_search_wolfe12(f, fprime, xk, pk, gfk, old_fval, old_old_fval,
                         **kwargs):
    """
    Same as line_search_wolfe1, but fall back to line_search_wolfe2 if
    suitable step length is not found, and raise an exception if a
    suitable step length is not found.
    Raises
    ------
    _LineSearchError
        If no suitable step size is found
    """
    ret = line_search_wolfe1(f, fprime, xk, pk, gfk,
                             old_fval, old_old_fval,
                             **kwargs)

    if ret[0] is None:
        # line search failed: try different one.
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', LineSearchWarning)
            ret = line_search_wolfe2(f, fprime, xk, pk, gfk,
                                     old_fval, old_old_fval)

    if ret[0] is None:
        raise _LineSearchError()

    return ret

def minimize_newtonlu(fun, x0, args=(), jac=None, hess=None, hessp=None,
                       callback=None, xtol=1e-5, eps=_epsilon, maxiter=None,
                       disp=False, return_all=False,
                       **unknown_options):
    """ Minimization of scalar function of one or more variables using the Newton-LU algorithm.

    .. note:: the `jac` parameter (Jacobian) is required.

    Args:
      disp (bool): Set to True to print convergence messages.
      xtol (float): Average relative error in solution `xopt` acceptable for
        convergence.
      maxiter (int): Maximum number of iterations to perform.
      eps (float or ndarray): If `jac` is approximated, use this value
        for the step size.
    """
    if jac is None:
        raise ValueError('Jacobian is required for Newton-LU method')
    f = fun
    fprime = jac
    fhess_p = hessp
    fhess = hess
    avextol = xtol
    epsilon = eps
    retall = return_all

    x0 = np.asarray(x0).flatten()
    fcalls, f = wrap_function(f, args)
    gcalls, fprime = wrap_function(fprime, args)
    hcalls = 0
    if maxiter is None:
        maxiter = len(x0)*200

    xtol = len(x0) * avextol
    update = [2 * xtol]
    xk = x0
    if retall:
        allvecs = [xk]
    k = 0
    old_fval = f(x0)
    old_old_fval = None
    float64eps = np.finfo(np.float64).eps
    warnflag = 0

    while (np.add.reduce(np.abs(update)) > xtol) and (k < maxiter):
        # Compute a search direction pk by applying the CG method to
        #  del2 f(xk) p = - grad f(xk) starting from 0.
        b = -fprime(xk)

        if fhess is not None:             # you want to compute hessian once.
            A = fhess(*(xk,) + args)
            hcalls = hcalls + 1
            #Now we need to check if the Hessian is positive definite.
            #If not, we will first try to modify it a certain number of times to make it positive definite.
            #After this number of trials we will revert to the steepest descent search direction
            Lchol, success, counterIt = makeHessianPositive_Cholesky( A )
            if success == True: #Newton direction
               b_temp = scialg.solve_triangular(Lchol, b, trans=0, lower=True)
               pk = scialg.solve_triangular(Lchol, b_temp, trans='T', lower=True) # search direction is solution to system.
            else: #Steepest descent
               pk = b

        gfk = -b    # gradient at xk

        try:
            alphak, fc, gc, old_fval, old_old_fval, gfkp1 = \
                     _line_search_wolfe12(f, fprime, xk, pk, gfk,
                                          old_fval, old_old_fval)
        except _LineSearchError:
            # Line search failed to find a better solution.
            warnflag = 2
            break

        update = alphak * pk
        xk = xk + update        # upcast if necessary
        if callback is not None:
            callback(xk)
        if retall:
            allvecs.append(xk)
        k += 1

    fval = old_fval
    if warnflag == 2:
        msg = _status_message['pr_loss']
        if disp:
            print("Warning: " + msg)
            print("         Current function value: %f" % fval)
            print("         Iterations: %d" % k)
            print("         Function evaluations: %d" % fcalls[0])
            print("         Gradient evaluations: %d" % gcalls[0])
            print("         Hessian evaluations: %d" % hcalls)
    elif k >= maxiter:
        warnflag = 1
        msg = _status_message['maxiter']
        if disp:
            print("Warning: " + msg)
            print("         Current function value: %f" % fval)
            print("         Iterations: %d" % k)
            print("         Function evaluations: %d" % fcalls[0])
            print("         Gradient evaluations: %d" % gcalls[0])
            print("         Hessian evaluations: %d" % hcalls)
    else:
        msg = _status_message['success']
        if disp:
            print(msg)
            print("         Current function value: %f" % fval)
            print("         Iterations: %d" % k)
            print("         Function evaluations: %d" % fcalls[0])
            print("         Gradient evaluations: %d" % gcalls[0])
            print("         Hessian evaluations: %d" % hcalls)

    result = OptimizeResult(fun=fval, jac=gfk, nfev=fcalls[0], njev=gcalls[0],
                            nhev=hcalls, status=warnflag,
                            success=(warnflag == 0), message=msg, x=xk,
                            nit=k)
    if retall:
        result['allvecs'] = allvecs
    return result


def minimize_newtoncglu(fun, x0, args=(), jac=None, hess=None, hessp=None,
                       callback=None, xtol=1e-5, eps=_epsilon, maxiter=None,
                       disp=False, return_all=False,
                       **unknown_options):
    """ Minimization of scalar function of one or more variables using the Newton-LU algorithm.

    .. note:: the `jac` parameter (Jacobian) is required.

    Args:
      disp (bool): Set to True to print convergence messages.
      xtol (float): Average relative error in solution `xopt` acceptable for
        convergence.
      maxiter (int): Maximum number of iterations to perform.
      eps (float or ndarray): If `jac` is approximated, use this value
        for the step size.
    """
    if jac is None:
        raise ValueError('Jacobian is required for Newton-LU method')
    f = fun
    fprime = jac
    fhess_p = hessp
    fhess = hess
    avextol = xtol
    epsilon = eps
    retall = return_all

    x0 = np.asarray(x0).flatten()
    fcalls, f = wrap_function(f, args)
    gcalls, fprime = wrap_function(fprime, args)
    hcalls = 0
    if maxiter is None:
        maxiter = len(x0)*200

    xtol = len(x0) * avextol
    update = [2 * xtol]
    xk = x0
    if retall:
        allvecs = [xk]
    k = 0
    old_fval = f(x0)
    old_old_fval = None
    float64eps = np.finfo(np.float64).eps
    warnflag = 0

    #### CG-ITERATION #####

    while (np.add.reduce(np.abs(update)) > xtol) and (k < maxiter):
        # Compute a search direction pk by applying the CG method to
        #  del2 f(xk) p = - grad f(xk) starting from 0.
        b = -fprime(xk)

        #Options iterative solver?
        maggrad = np.add.reduce(np.abs(b))
        eta = np.min([0.5, np.sqrt(maggrad)])
        termcond = eta * maggrad
        xsupi = np.zeros(len(x0), dtype=x0.dtype)
        ri = -b
        psupi = -ri
        i = 0
        dri0 = np.dot(ri, ri)

        if fhess is not None:             # you want to compute hessian once.
            A = fhess(*(xk,) + args)
            hcalls = hcalls + 1

        while np.add.reduce(np.abs(ri)) > termcond:
            if fhess is None:
                if fhess_p is None:
                    Ap = approx_fhess_p(xk, psupi, fprime, epsilon)
                else:
                    Ap = fhess_p(xk, psupi, *args)
                    hcalls = hcalls + 1
            else:
                Ap = np.dot(A, psupi)
            # check curvature
            Ap = np.asarray(Ap).squeeze()  # get rid of matrices...
            curv = np.dot(psupi, Ap)
            if 0 <= curv <= 3 * float64eps:
                break
            elif curv < 0:
                if (i > 0):
                    break
                else:
                    # fall back to steepest descent direction
                    xsupi = dri0 / (-curv) * b
                    break
            alphai = dri0 / curv
            xsupi = xsupi + alphai * psupi
            ri = ri + alphai * Ap
            dri1 = np.dot(ri, ri)
            betai = dri1 / dri0
            psupi = -ri + betai * psupi
            i = i + 1
            dri0 = dri1          # update np.dot(ri,ri) for next time.

        pk = xsupi  # search direction is solution to system.
        gfk = -b    # gradient at xk

        try:
            alphak, fc, gc, old_fval, old_old_fval, gfkp1 = \
                     _line_search_wolfe12(f, fprime, xk, pk, gfk,
                                          old_fval, old_old_fval)
        except _LineSearchError:
            # Line search failed to find a better solution.
            warnflag = 2
            break

        update = alphak * pk
        xk = xk + update        # upcast if necessary
        if callback is not None:
            callback(xk)
        if retall:
            allvecs.append(xk)
        k += 1

    #### EXACT LU-ITERATION #####
    update = [2 * xtol]
    while (np.add.reduce(np.abs(update)) > xtol) and (k < maxiter):
        # Compute a search direction pk by applying the CG method to
        #  del2 f(xk) p = - grad f(xk) starting from 0.
        b = -fprime(xk)

        if fhess is not None:             # you want to compute hessian once.
            A = fhess(*(xk,) + args)
            hcalls = hcalls + 1
            #Now we need to check if the Hessian is positive definite.
            #If not, we will first try to modify it a certain number of times to make it positive definite.
            #After this number of trials we will revert to the steepest descent search direction
            Lchol, success, counterIt = makeHessianPositive_Cholesky( A )
            if success == True: #Newton direction
               b_temp = scialg.solve_triangular(Lchol, b, trans=0, lower=True)
               pk = scialg.solve_triangular(Lchol, b_temp, trans='T', lower=True) # search direction is solution to system.
            else: #Steepest descent
               pk = b

        gfk = -b    # gradient at xk

        try:
            alphak, fc, gc, old_fval, old_old_fval, gfkp1 = \
                     _line_search_wolfe12(f, fprime, xk, pk, gfk,
                                          old_fval, old_old_fval)
        except _LineSearchError:
            # Line search failed to find a better solution.
            warnflag = 2
            break

        update = alphak * pk
        xk = xk + update        # upcast if necessary
        if callback is not None:
            callback(xk)
        if retall:
            allvecs.append(xk)
        k += 1


    fval = old_fval
    if warnflag == 2:
        msg = _status_message['pr_loss']
        if disp:
            print("Warning: " + msg)
            print("         Current function value: %f" % fval)
            print("         Iterations: %d" % k)
            print("         Function evaluations: %d" % fcalls[0])
            print("         Gradient evaluations: %d" % gcalls[0])
            print("         Hessian evaluations: %d" % hcalls)
    elif k >= maxiter:
        warnflag = 1
        msg = _status_message['maxiter']
        if disp:
            print("Warning: " + msg)
            print("         Current function value: %f" % fval)
            print("         Iterations: %d" % k)
            print("         Function evaluations: %d" % fcalls[0])
            print("         Gradient evaluations: %d" % gcalls[0])
            print("         Hessian evaluations: %d" % hcalls)
    else:
        msg = _status_message['success']
        if disp:
            print(msg)
            print("         Current function value: %f" % fval)
            print("         Iterations: %d" % k)
            print("         Function evaluations: %d" % fcalls[0])
            print("         Gradient evaluations: %d" % gcalls[0])
            print("         Hessian evaluations: %d" % hcalls)

    result = OptimizeResult(fun=fval, jac=gfk, nfev=fcalls[0], njev=gcalls[0],
                            nhev=hcalls, status=warnflag,
                            success=(warnflag == 0), message=msg, x=xk,
                            nit=k)
    if retall:
        result['allvecs'] = allvecs
    return result




##################### Auxiliary functions #############################################

def wrap_function(function, args):
    ncalls = [0]
    if function is None:
        return ncalls, None

    def function_wrapper(*wrapper_args):
        ncalls[0] += 1
        return function(*(wrapper_args + args))

    return ncalls, function_wrapper


def makeHessianPositive_Cholesky( hess, beta = 1e-3 , maxIt = 10, prodFactor = 2 ):
  # From "Numerical Optimization" Nocedal & Wright pg.51

  if isSymMatrix( hess )==False:
     hess = closestSymMatrix( hess ) # Force symmetry
  diagHess = np.diag(hess)
  dim = len(diagHess)
  hess_temp = np.copy(hess)

  #Decide initial shift
  if np.min(diagHess) > 0:
     tau = 0
  else:
     tau = - np.min(diagHess) + beta

  counterIt = 0
  flag = False

  while  (flag == False and counterIt<maxIt):
    hess_temp[:,:] = hess  # Assign original hessian
    diag_view = np.diagonal(hess_temp)
    diag_view.setflags(write=True)
    diag_view += tau  # hess + tau*Id


    try:
      Lchol = np.linalg.cholesky(hess_temp)
      flag = True
    except LinAlgError:
      pass

    tau = max( prodFactor*tau, beta )
    counterIt += 1 # Update counter

    if flag == True:
      success = True
    else:
      success = False
      Lchol=0

  return Lchol, success, counterIt-1

def closestSymMatrix( A ):
  # Override the square matrix A with the closest symmetric matrix in the Frobenius norm(A+A^T) /2
  return (A+ np.transpose(A))/2


def isSymMatrix( A ):
  return (A.transpose() == A).all()





#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

from TransportMaps import deprecate
from TransportMaps.Distributions import *

__all__ = ["Density", "ParametricDensity",
           "TransportMapDensity",
           "PushForwardTransportMapDensity",
           "PullBackTransportMapDensity",
           'FrozenDensity_1d',
           'GaussianDensity', 'StandardNormalDensity',
           'LogNormalDensity', 'LogisticDensity',
           'GammaDensity', 'BetaDensity',
           'WeibullDensity',
           'GumbelDensity', 'BananaDensity']

class Density(Distribution):
    @deprecate("TransportMaps.Densities.Density", "1.0",
               "Use TransportMaps.Distributions.Distribution instead")
    def __init__(self, *args, **kwargs):
        super(Density, self).__init__(*args, **kwargs)

class ParametricDensity(ParametricDistribution):
    @deprecate("TransportMaps.Densities.ParametricDensity", "1.0",
               "Use TransportMaps.Distributions.ParametricDistribution instead")
    def __init__(self, *args, **kwargs):
        super(ParametricDensity, self).__init__(*args, **kwargs)

class TransportMapDensity(TransportMapDistribution):
    @deprecate("TransportMaps.Densities.TransportMapDensity", "1.0",
               "Use TransportMaps.Distributions.TransportMapDistribution instead")
    def __init__(self, *args, **kwargs):
        super(TransportMapDensity, self).__init__(*args, **kwargs)

    @property
    def base_density(self):
        return self.base_distribution

    @base_density.setter
    def base_density(self, value):
        self.base_distribution = value

class PushForwardTransportMapDensity(PushForwardTransportMapDistribution):
    @deprecate(
        "TransportMaps.Densities.PushForwardTransportMapDensity", "1.0",
        "Use TransportMaps.Distributions.PushForwardTransportMapDistribution instead")
    def __init__(self, *args, **kwargs):
        super(PushForwardTransportMapDensity, self).__init__(*args, **kwargs)

    @property
    def base_density(self):
        return self.base_distribution

    @base_density.setter
    def base_density(self, value):
        self.base_distribution = value

class PullBackTransportMapDensity(PullBackTransportMapDistribution):
    @deprecate(
        "TransportMaps.Densities.PullBackTransportMapDensity", "1.0",
        "Use TransportMaps.Distributions.PullBackTransportMapDistribution instead")
    def __init__(self, *args, **kwargs):
        super(PullBackTransportMapDensity, self).__init__(*args, **kwargs)

    @property
    def base_density(self):
        return self.base_distribution

    @base_density.setter
    def base_density(self, value):
        self.base_distribution = value

# FROZEN DISTRIBUTIONS
class FrozenDensity_1d(FrozenDistribution_1d):
    @deprecate("TransportMaps.Densities.FrozenDensity_1d", "1.0",
               "Use TransportMaps.Distributions.FrozenDistribution_1d instead")
    def __init__(self, *args, **kwargs):
        super(FrozenDensity_1d, self).__init__(*args, **kwargs)

class GaussianDensity(GaussianDistribution):
    @deprecate("TransportMaps.Densities.GaussianDensity", "1.0",
               "Use TransportMaps.Distributions.GaussianDistribution instead")
    def __init__(self, *args, **kwargs):
        super(GaussianDensity, self).__init__(*args, **kwargs)

class StandardNormalDensity(StandardNormalDistribution):
    @deprecate("TransportMaps.Densities.StandardNormalDensity", "1.0",
               "Use TransportMaps.Distributions.StandardNormalDistribution instead")
    def __init__(self, *args, **kwargs):
        super(StandardNormalDensity, self).__init__(*args, **kwargs)

class LogNormalDensity(LogNormalDistribution):
    @deprecate("TransportMaps.Densities.LogNormalDensity", "1.0",
               "Use TransportMaps.Distributions.LogNormalDistribution instead")
    def __init__(self, *args, **kwargs):
        super(LogNormalDensity, self).__init__(*args, **kwargs)

class LogisticDensity(LogisticDistribution):
    @deprecate("TransportMaps.Densities.LogisticDensity", "1.0",
               "Use TransportMaps.Distributions.LogisticDistribution instead")
    def __init__(self, *args, **kwargs):
        super(LogisticDensity, self).__init__(*args, **kwargs)

class GammaDensity(GammaDistribution):
    @deprecate("TransportMaps.Densities.GammaDensity", "1.0",
               "Use TransportMaps.Distributions.GammaDistribution instead")
    def __init__(self, *args, **kwargs):
        super(GammaDensity, self).__init__(*args, **kwargs)

class BetaDensity(BetaDistribution):
    @deprecate("TransportMaps.Densities.BetaDensity", "1.0",
               "Use TransportMaps.Distributions.BetaDistribution instead")
    def __init__(self, *args, **kwargs):
        super(BetaDensity, self).__init__(*args, **kwargs)

class WeibullDensity(WeibullDistribution):
    @deprecate("TransportMaps.Densities.WeibullDensity", "1.0",
               "Use TransportMaps.Distributions.WeibullDistribution instead")
    def __init__(self, *args, **kwargs):
        super(WeibullDensity, self).__init__(*args, **kwargs)

class GumbelDensity(GumbelDistribution):
    @deprecate("TransportMaps.Densities.GumbelDensity", "1.0",
               "Use TransportMaps.Distributions.GumbelDistribution instead")
    def __init__(self, *args, **kwargs):
        super(GumbelDensity, self).__init__(*args, **kwargs)

class BananaDensity(BananaDistribution):
    @deprecate("TransportMaps.Densities.BananaDensity", "1.0",
               "Use TransportMaps.Distributions.BananaDistribution instead")
    def __init__(self, *args, **kwargs):
        super(BananaDensity, self).__init__(*args, **kwargs)

